import morepath
import onegov.core.security

from onegov.core.framework import Framework
from onegov.core.security import Public, Personal, Private, Secret
from onegov.core.security import forget, remembered
from webtest import TestApp as Client


def spawn_basic_permissions_app(redis_url):

    class App(Framework):
        pass

    class Root(object):
        pass

    @App.path(path='', model=Root)
    def get_root():
        return Root()

    @App.path(path='/hidden')
    class HiddenFromPublic(object):
        is_hidden_from_public = True

    @App.view(model=Root, name='public', permission=Public)
    def public_view(self, request):
        return 'public'

    @App.view(model=Root, name='private', permission=Private)
    def private_view(self, request):
        return 'private'

    @App.view(model=Root, name='personal', permission=Personal)
    def personal_view(self, request):
        return 'personal'

    @App.view(model=Root, name='secret', permission=Secret)
    def secret_view(self, request):
        return 'secret'

    @App.view(model=HiddenFromPublic, permission=Public)
    def hidden_from_public_view(self, request):
        return 'hidden'

    @App.view(
        model=Root,
        name='login',
        permission=Public,
        request_method='POST'
    )
    def login(self, request):
        userid = request.params.get('userid')
        role = request.params.get('role')
        identity = request.app.application_bound_identity(userid, role)

        @request.after
        def remember(response):
            request.app.remember_identity(response, request, identity)

    @App.view(model=Root, name='logout', permission=Private)
    def logout(self, request):

        @request.after
        def forget(response):
            request.app.forget_identity(response, request)

    # the scan is required (happens automatically if using onegov.server)
    morepath.scan(onegov.core.security)

    app = App()
    app.namespace = 'test'
    app.configure_application(identity_secure=False, redis_url=redis_url)
    app.set_application_id('test/app')

    return app


def test_anonymous_access(redis_url):
    client = Client(spawn_basic_permissions_app(redis_url))

    assert client.get('/public').text == 'public'
    assert client.get('/personal', expect_errors=True).status_code == 403
    assert client.get('/private', expect_errors=True).status_code == 403
    assert client.get('/secret', expect_errors=True).status_code == 403
    assert client.get('/logout', expect_errors=True).status_code == 403
    assert client.get('/hidden', expect_errors=True).status_code == 403


def test_personal_access(redis_url):
    client = Client(spawn_basic_permissions_app(redis_url))
    # use the userid 'admin' to be sure that we don't let it matter
    client.post('/login', {'userid': 'admin', 'role': 'member'})

    assert client.get('/public').text == 'public'
    assert client.get('/personal').text == 'personal'
    assert client.get('/private', expect_errors=True).status_code == 403
    assert client.get('/secret', expect_errors=True).status_code == 403
    assert client.get('/logout', expect_errors=True).status_code == 403
    assert client.get('/hidden', expect_errors=True).status_code == 403


def test_private_access(redis_url):
    client = Client(spawn_basic_permissions_app(redis_url))
    # use the userid 'admin' to be sure that we don't let it matter
    client.post('/login', {'userid': 'admin', 'role': 'editor'})

    assert client.get('/public').text == 'public'
    assert client.get('/personal').text == 'personal'
    assert client.get('/private').text == 'private'
    assert client.get('/hidden').text == 'hidden'
    assert client.get('/secret', expect_errors=True).status_code == 403

    client.get('/logout')

    assert client.get('/public').text == 'public'
    assert client.get('/personal', expect_errors=True).status_code == 403
    assert client.get('/private', expect_errors=True).status_code == 403
    assert client.get('/secret', expect_errors=True).status_code == 403
    assert client.get('/logout', expect_errors=True).status_code == 403
    assert client.get('/hidden', expect_errors=True).status_code == 403


def test_secret_access(redis_url):
    client = Client(spawn_basic_permissions_app(redis_url))
    # use the userid 'editor' to be sure that we don't let it matter
    client.post('/login', {'userid': 'editor', 'role': 'admin'})

    assert client.get('/public').text == 'public'
    assert client.get('/personal').text == 'personal'
    assert client.get('/private').text == 'private'
    assert client.get('/secret').text == 'secret'
    assert client.get('/hidden').text == 'hidden'

    client.get('/logout')

    assert client.get('/public').text == 'public'
    assert client.get('/personal', expect_errors=True).status_code == 403
    assert client.get('/private', expect_errors=True).status_code == 403
    assert client.get('/secret', expect_errors=True).status_code == 403
    assert client.get('/logout', expect_errors=True).status_code == 403
    assert client.get('/hidden', expect_errors=True).status_code == 403


def test_secure_cookie(redis_url):
    app = spawn_basic_permissions_app(redis_url)
    app.identity_secure = True

    client = Client(app)
    client.post(
        '/login', {'userid': 'editor', 'role': 'admin'},
        extra_environ={'wsgi.url_scheme': 'http'}
    )

    cookie = client.cookiejar._cookies['localhost.local']['/']['session_id']
    assert cookie.secure


def test_forget(redis_url):
    app = spawn_basic_permissions_app(redis_url)
    client = Client(app)
    response = client.post('/login', {'userid': 'user', 'role': 'admin'})

    session_id = app.unsign(response.request.cookies['session_id'])
    assert remembered(app, session_id)

    assert client.get('/public').text == 'public'
    assert client.get('/personal').text == 'personal'
    assert client.get('/private').text == 'private'
    assert client.get('/secret').text == 'secret'
    assert client.get('/hidden').text == 'hidden'

    forget(app, session_id)
    assert not remembered(app, session_id)

    assert client.get('/public').text == 'public'
    assert client.get('/personal', expect_errors=True).status_code == 403
    assert client.get('/private', expect_errors=True).status_code == 403
    assert client.get('/secret', expect_errors=True).status_code == 403
    assert client.get('/logout', expect_errors=True).status_code == 403
    assert client.get('/hidden', expect_errors=True).status_code == 403



import psycopg2

from onegov.core.orm.cache import orm_cached
from onegov.core.orm.session_manager import SessionManager, query_schemas
from onegov.core.orm.sql import as_selectable, as_selectable_from_path
from sqlalchemy import event, inspect
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import object_session
from sqlalchemy.orm import Query
from sqlalchemy_utils import TranslationHybrid
from zope.sqlalchemy import mark_changed
from sqlalchemy.exc import InterfaceError, OperationalError


MISSING = object()

DB_CONNECTION_ERRORS = (
    OperationalError,
    InterfaceError,
    psycopg2.OperationalError,
    psycopg2.InterfaceError,
)


#: The base for all OneGov Core ORM Models
class ModelBase(object):

    #: set by :class:`onegov.core.orm.cache.OrmCacheDescriptor`, this attribute
    #: indicates if the current model was loaded from cache
    is_cached = False

    @classmethod
    def get_polymorphic_class(cls, identity_value, default=MISSING):
        """ Returns the polymorphic class if it exists, given the value
        of the polymorphic identity.

        Asserts that the identity is actually found, unless a default is
        provided.

        """
        mapper = inspect(cls).polymorphic_map.get(identity_value)

        if default is MISSING:
            assert mapper, "No such polymorphic_identity: {}".format(
                identity_value
            )

        return mapper and mapper.class_ or default

    @property
    def session_manager(self):
        return SessionManager.get_active()


Base = declarative_base(cls=ModelBase)

#: A translation hybrid integrated with OneGov Core. See also:
#: http://sqlalchemy-utils.readthedocs.org/en/latest/internationalization.html
translation_hybrid = TranslationHybrid(
    current_locale=lambda obj: obj.session_manager.current_locale,
    default_locale=lambda obj: obj.session_manager.default_locale,
)


def find_models(base, is_match):
    """ Finds the ORM models in the given ORM base class that match a filter.

    The filter is called with each class in the instance and it is supposed
    to return True if it matches.

    For example, find all SQLAlchemy models that use
    :class:`~onegov.core.orm.mixins.ContentMixin`::

        from onegov.core.orm.mixins import ContentMixin
        find_models(base, is_match=lambda cls: issubclass(cls, ContentMixin))

    """
    for cls in base.__subclasses__():
        if is_match(cls):
            yield cls

        for cls in find_models(cls, is_match):
            yield cls


def configure_listener(cls, key, instance):
    """ The zope.sqlalchemy transaction mechanism doesn't recognize changes to
    cached objects. The following code intercepts all object changes and marks
    the transaction as changed if there was a change to a cached object.

    """

    def mark_as_changed(obj, *args, **kwargs):
        if obj.is_cached:
            mark_changed(object_session(obj))

    event.listen(instance, 'append', mark_as_changed)
    event.listen(instance, 'remove', mark_as_changed)
    event.listen(instance, 'set', mark_as_changed)
    event.listen(instance, 'init_collection', mark_as_changed)
    event.listen(instance, 'dispose_collection', mark_as_changed)


event.listen(ModelBase, 'attribute_instrument', configure_listener)


def share_session_manager(query):
    session_manager = SessionManager.get_active()

    for desc in query.column_descriptions:
        desc['type'].session_manager = session_manager


event.listen(Query, 'before_compile', share_session_manager, retval=False)


__all__ = [
    'Base',
    'SessionManager',
    'as_selectable',
    'as_selectable_from_path',
    'translation_hybrid',
    'find_models',
    'orm_cached',
    'query_schemas'
]



import re

from bleach.sanitizer import Cleaner
from html2text import HTML2Text

# html tags allowed by bleach
SANE_HTML_TAGS = [
    'a',
    'abbr',
    'b',
    'br',
    'blockquote',
    'code',
    'del',
    'div',
    'em',
    'i',
    'img',
    'hr',
    'li',
    'ol',
    'p',
    'pre',
    'strong',
    'sup',
    'span',
    'ul',
    'h1',
    'h2',
    'h3',
    'h4',
    'h5',
    'h6'
]

# html attributes allowed by bleach
SANE_HTML_ATTRS = {
    'a': ['href', 'title'],
    'abbr': ['title', ],
    'acronym': ['title', ],
    'img': ['src', 'alt', 'title']
}

# lines without these plaintext characters are excluded in html_to_text
VALID_PLAINTEXT_CHARACTERS = re.compile(r"""
    [
        \d  # decimals
        \w  # words
        \n  # new lines

        # emojis
        \U00002600-\U000027BF
        \U0001f300-\U0001f64F
        \U0001f680-\U0001f6FF
    ]+
""", re.VERBOSE)

# match empty link expressions
EMPTY_LINK = re.compile(r"\[\]\([^)]+\)")


cleaner = Cleaner(
    tags=SANE_HTML_TAGS,
    attributes=SANE_HTML_ATTRS
)


def sanitize_html(html):
    """ Takes the given html and strips all but a whitelisted number of tags
    from it.

    """

    return cleaner.clean(html)


def sanitize_svg(svg):
    """ I couldn't find a good svg sanitiser function yet, so for now
    this function will be a no-op, though it will try to detect
    svg files which are harmful.

    I tried to go with bleach/html5lib, but the lack of xml namespace support
    makes those options a no go.

    In the future we want a proper SVG sanitiser here!

    """

    assert 'javascript:' not in svg
    assert 'CDATA' not in svg
    assert '<script>' not in svg
    assert 'Set-Cookie' not in svg

    return svg


def html_to_text(html, **config):
    """ Takes the given HTML text and extracts the text from it.

    The result is markdown. The driver behind it is html2text. Have a look
    at https://github.com/Alir3z4/html2text/blob/master/html2text/__init__.py
    to see all options.

    """

    # filter out duplicated lines and lines without any text
    html2text = HTML2Text()

    # output unicode directly, instead of approximating it to ASCII
    config.setdefault('unicode_snob', True)

    # do not wrap lines after a certain length
    config.setdefault('body_width', 0)

    # images are just converted into somewhat useless links, so disable
    config.setdefault('ignore_images', True)

    # we do our own paragraph handling
    config.setdefault('single_line_break', True)

    for key, value in config.items():
        setattr(html2text, key, value)

    lines = html2text.handle(html).splitlines()

    # ignore images doesn't catch all images:
    if config['ignore_images']:
        lines = (EMPTY_LINK.sub('', line) for line in lines)

    lines = (l.strip() for l in lines)
    lines = (l for l in lines if VALID_PLAINTEXT_CHARACTERS.search(l))

    # use double newlines to get paragraphs
    plaintext = '\n\n'.join(lines)

    # in an attempt to create proper markdown html2text will escape
    # dots. Since markdown is not something we care about here, we undo that
    plaintext = plaintext.replace('\\.', '.')

    return plaintext



from onegov.core.cronjobs import Job
from onegov.core.framework import Framework
from onegov.core.security import Public
from webob.exc import HTTPException


@Framework.permission_rule(model=object, permission=object, identity=None)
def has_permission_not_logged_in(app, identity, model, permission):
    """ This catch-all rule returns the default permission rule. It says
    that the permission must be part of the anonymous rule.

    Models with a 'is_hidden_from_public' property are not viewable by
    anonymous users, if said property is set to True. If it is set to False,
    the usual permission check kicks in.

    """

    if getattr(model, 'is_hidden_from_public', False):
        return False

    return permission in app.settings.roles.anonymous


@Framework.permission_rule(model=object, permission=object)
def has_permission_logged_in(app, identity, model, permission):
    """ This permission rule matches all logged in identities. It requires
    the identity to have a 'role' attribute. Said role attribute is used
    to determine if the given permission is part of the given role.

    """

    assert hasattr(identity, 'role'), """
        the identity needs to implement a role property
    """

    # basic members do not get access to hidden from public
    if identity.role == 'member':
        if getattr(model, 'is_hidden_from_public', False):
            return False

    return permission in getattr(app.settings.roles, identity.role)


@Framework.permission_rule(
    model=HTTPException,
    permission=Public,
    identity=None)
def may_view_http_errors_not_logged_in(app, identity, model, permission):
    """ HTTP errors may be viewed by anyone, regardeless of settings.

    This is important, otherwise the HTTPForbidden/HTTPNotFound views
    will lead to an exception if the user does not have the ``Public``
    permission.

    """
    return True


@Framework.permission_rule(
    model=Job,
    permission=Public,
    identity=None)
def may_view_cronjobs_not_logged_in(app, identity, model, permission):
    """ Cronjobs are run anonymously from a thread and need to be excluded
    from the permission rules as a result.

    """
    return True



import morepath
import pytest

from datetime import datetime, timedelta
from freezegun import freeze_time
from onegov.core.framework import Framework
from onegov.core.request import CoreRequest, ReturnToMixin
from onegov.core.security import Public, Personal, Private, Secret
from onegov.core.utils import Bunch, scan_morepath_modules
from webtest import TestApp as Client
from urllib.parse import quote


def test_url_safe_token():

    request = CoreRequest(environ={
        'PATH_INFO': '/',
        'SERVER_NAME': '',
        'SERVER_PORT': '',
        'SERVER_PROTOCOL': 'https'
    }, app=Bunch(identity_secret='asdf', lookup=None))

    token = request.new_url_safe_token({'foo': 'bar'})

    assert request.load_url_safe_token(token) == {'foo': 'bar'}
    assert request.load_url_safe_token(token, salt='x') is None

    with freeze_time(datetime.now() + timedelta(seconds=2)):
        assert request.load_url_safe_token(token, max_age=1) is None


def test_return_to_mixin():

    class Request(ReturnToMixin):

        GET = {}

        @property
        def identity_secret(self):
            return 'foobar'

        @property
        def url(self):
            return 'http://here'

    def param(url):
        return url.split('=')[1]

    r = Request()

    assert r.return_to('https://example.org', '/')\
        .startswith('https://example.org?return-to=')

    assert r.return_here('https://example.org')\
        .startswith('https://example.org?return-to=')

    r.GET['return-to'] = 'http://phising'
    assert r.redirect('http://safe').location == 'http://safe'

    r.GET['return-to'] = param(r.return_here('http://safe'))
    assert r.redirect('http://safe').location == 'http://here'

    r.GET['return-to'] = param(r.return_to('http://safe', 'http://known'))
    assert r.redirect('http://safe').location == 'http://known'

    r.GET['return-to'] = param(r.return_to('http://safe', 'http://known'))
    r.GET['return-to'] += 'tampering'
    assert r.redirect('http://safe').location == 'http://safe'


def test_return_to(redis_url):

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.view(model=Root)
    def homepage(self, request):
        return request.return_to(request.link(self, 'do-something'), '/')

    @App.view(model=Root, name='do-something')
    def do_something(self, request):
        return request.redirect('/default')

    App.commit()

    app = App()
    app.application_id = 'test'
    app.configure_application(identity_secure=False, redis_url=redis_url)

    c = Client(app)
    do_something_url = c.get('/').text
    assert c.get(do_something_url).location == 'http://localhost/'
    assert c.get(do_something_url + 'x').location == 'http://localhost/default'
    assert c.get('/do-something').location == 'http://localhost/default'


def test_has_permission(redis_url):

    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        allowed_for = (Public, Personal, Private, Secret)

    @App.view(model=Root, permission=Public)
    def view(self, request):
        permission = {
            'public': Public,
            'personal': Personal,
            'private': Private,
            'secret': Secret
        }[request.params.get('permission')]

        return request.has_permission(self, permission) and 'true' or 'false'

    @App.view(model=Root, permission=Public, name='login')
    def login(self, request):

        @request.after
        def remember_identity(response):
            request.app.remember_identity(response, request, morepath.Identity(
                userid='foo',
                role='admin',
                application_id=request.app.application_id
            ))

        return 'ok'

    @App.permission_rule(model=Root, permission=object, identity=None)
    def has_permission_not_logged_in(identity, model, permission):
        return permission is Public

    @App.permission_rule(model=Root, permission=object)
    def has_permission_logged_in(identity, model, permission):
        return permission in model.allowed_for

    scan_morepath_modules(App)
    App.commit()

    app = App()

    app.application_id = 'test'
    app.configure_application(identity_secure=False, redis_url=redis_url)

    c = Client(app)
    assert c.get('/?permission=public').text == 'true'
    assert c.get('/?permission=personal').text == 'false'
    assert c.get('/?permission=private').text == 'false'
    assert c.get('/?permission=secret').text == 'false'

    c.get('/login')

    assert c.get('/?permission=public').text == 'true'
    assert c.get('/?permission=personal').text == 'true'
    assert c.get('/?permission=private').text == 'true'
    assert c.get('/?permission=secret').text == 'true'


def test_permission_by_view(redis_url):
    class App(Framework):
        pass

    @App.path(path='/')
    class Root(object):
        pass

    @App.view(model=Root, name='public', permission=Public)
    def public_view(self, request):
        assert False  # we don't want this view to be called

    @App.view(model=Root, name='personal', permission=Personal)
    def personal_view(self, request):
        assert False  # we don't want this view to be called

    @App.view(model=Root, name='private', permission=Private)
    def private_view(self, request):
        assert False  # we don't want this view to be called

    @App.view(model=Root, name='secret', permission=Secret)
    def secret_view(self, request):
        assert False  # we don't want this view to be called

    @App.view(model=Root, name='login', permission=Public)
    def login(self, request):

        @request.after
        def remember_identity(response):
            request.app.remember_identity(response, request, morepath.Identity(
                userid='foo',
                role='admin',
                application_id=request.app.application_id
            ))

        return 'ok'

    @App.view(model=Root)
    def view(self, request):
        url = request.params.get('url')
        return request.has_access_to_url(url) and 'true' or 'false'

    scan_morepath_modules(App)
    App.commit()

    app = App()

    app.application_id = 'test'
    app.configure_application(identity_secure=False, redis_url=redis_url)

    c = Client(app)

    def has_access(url):
        return c.get('/?url={}'.format(quote(url))).text == 'true'

    assert app.permission_by_view(Root) is None
    assert app.permission_by_view(Root, 'login') is Public
    assert app.permission_by_view(Root, 'public') is Public
    assert app.permission_by_view(Root, 'personal') is Personal
    assert app.permission_by_view(Root, 'private') is Private
    assert app.permission_by_view(Root, 'secret') is Secret

    assert app.permission_by_view(Root()) is None
    assert app.permission_by_view(Root(), 'login') is Public
    assert app.permission_by_view(Root(), 'public') is Public
    assert app.permission_by_view(Root(), 'personal') is Personal
    assert app.permission_by_view(Root(), 'private') is Private
    assert app.permission_by_view(Root(), 'secret') is Secret

    with pytest.raises(AssertionError):
        app.permission_by_view(None)

    with pytest.raises(KeyError):
        app.permission_by_view(Root, 'foobar')

    # the domain is ignored
    for domain in ('', 'https://example.org'):
        assert has_access('{}/'.format(domain))
        assert has_access('{}/login'.format(domain))
        assert has_access('{}/public'.format(domain))
        assert not has_access('{}/personal'.format(domain))
        assert not has_access('{}/private'.format(domain))
        assert not has_access('{}/secret'.format(domain))

    c.get('/login')

    for domain in ('', 'https://example.org'):
        assert has_access('{}/'.format(domain))
        assert has_access('{}/login'.format(domain))
        assert has_access('{}/public'.format(domain))
        assert has_access('{}/personal'.format(domain))
        assert has_access('{}/private'.format(domain))
        assert has_access('{}/secret'.format(domain))



from sedate import standardize_date, to_timezone
from sqlalchemy import types


class UTCDateTime(types.TypeDecorator):
    """ Stores dates as UTC.

    Internally, they are stored as timezone naive, because Postgres takes
    the local timezone into account when working with timezones. Values taken
    and values returned are forced to be timezone-aware though.

    """

    impl = types.DateTime

    def __init__(self):
        super().__init__(timezone=False)

    def process_bind_param(self, value, engine):
        if value is not None:
            return to_timezone(value, 'UTC').replace(tzinfo=None)

    def process_result_value(self, value, engine):
        if value is not None:
            return standardize_date(value, timezone='UTC')



from onegov.core.markdown import render_untrusted_markdown as render
from mistletoe import HTMLRenderer


def test_render_untrusted_markdown():

    # no unsafe html
    assert render("<table></table>") == '&lt;table&gt;&lt;/table&gt;\n'

    # this would usually result ina code class element, but we currently
    # do not allow classes to be set by untrusted html
    assert render('```python\nfoo\n```') == '<pre><code>foo\n</code></pre>\n'

    # inline html is disabled by default
    assert render('<a>foo</a>') == '<p>&lt;a&gt;foo&lt;/a&gt;</p>\n'

    # though we can enable it (it is still sanitized though)
    assert render('<a onclick="alert">foo</a>', cls=HTMLRenderer)\
        == '<p><a>foo</a></p>\n'


def test_markdown_line_endings():
    assert render('foo  \r\nbar') == render('foo  \nbar')



from chameleon import PageTemplate
from onegov.core import cache
from onegov.core.framework import Framework
from onegov.core.utils import Bunch


CALL_COUNT = 0


# cannot be pickled by the builtin pickler
class Point(object):

    __slots__ = ('x', 'y')

    def __init__(self, x, y):
        self.x = x
        self.y = y


def test_cache_key(redis_url):
    region = cache.get(namespace='ns', expiration_time=60, redis_url=redis_url)
    region.set('x' * 500, 'y')  # used to fail on the old memcached system


def test_cache_page_template(redis_url):
    region = cache.get(namespace='ns', expiration_time=60, redis_url=redis_url)
    region.set('template', PageTemplate('<!DOCTYPE html>'))
    region.get('template')


def test_redis(redis_url):
    app = Framework()
    app.namespace = 'towns'
    app.set_application_id('towns/detroit')
    app.configure_application(redis_url=redis_url)
    app.cache.set('foobar', Bunch(foo='bar'))

    assert app.cache.get('foobar').foo == 'bar'


def test_store_slots_redis(redis_url):
    # the following fails without the usage of an advanced pickler
    app = Framework()
    app.namespace = 'towns'
    app.set_application_id('towns/washington')

    app.configure_application(redis_url=redis_url)
    app.cache.set('point', Point(0, 0))
    assert app.cache.get('point').x == 0
    assert app.cache.get('point').y == 0


def test_cache_independence(redis_url):
    app = Framework()
    app.namespace = 'towns'
    app.set_application_id('towns/washington')

    app.configure_application(redis_url=redis_url)
    app.cache.set('foo', 'bar')
    assert app.cache.get('foo')

    app.set_application_id('towns/newyork')
    assert not app.cache.get('foo')

    app.namespace = 'cities'
    app.set_application_id('cities/washington')
    assert not app.cache.get('foo')

    app.namespace = 'towns'
    app.set_application_id('towns/washington')
    assert app.cache.get('foo')



import pytest

from datetime import datetime
from decimal import Decimal
from onegov.core.custom import json


class Point(object):

    def __init__(self, x, y):
        self.x = x
        self.y = y

    def __eq__(self, other):
        return self.x == other.x and self.y == other.y


def test_custom_json():

    dt = datetime(2015, 6, 25, 12, 0)

    data = {
        'datetime': dt,
        'date': dt.date(),
        'time': dt.time(),
        'generator': (x for x in range(0, 4)),
        'decimal': Decimal('3.1415926')
    }

    text = json.dumps(data)

    assert '__time__@12:00:00' in text
    assert '__date__@2015-06-25' in text
    assert '__datetime__@2015-06-25T12:00:00' in text
    assert '__decimal__@3.1415926' in text
    assert '[0,1,2,3]' in text

    data['generator'] = [x for x in range(0, 4)]
    assert json.loads(text) == data


def test_not_serializable():
    with pytest.raises(TypeError):
        json.dumps({'x': object()})


def test_sort_keys():
    data = {'c': 1, 'a': 2}
    assert json.dumps(data, sort_keys=True) == '{"a":2,"c":1}'


def test_prefix_serializer():
    prefix = json.PrefixSerializer(
        target=str,
        prefix='upper',
        encode=lambda s: s.upper(),
        decode=lambda s: s.lower()
    )

    assert prefix.encode('test') == '__upper__@TEST'
    assert prefix.decode('__upper__@TEST') == 'test'


def test_dictionary_serializer():

    d = json.DictionarySerializer(
        target=Point,
        keys=('x', 'y')
    )

    d.encode(Point(1, 2)) == {'x': 1, 'y': 2}
    d.decode({'x': 1, 'y': 2}) == Point(1, 2)

    with pytest.raises(TypeError):
        d.decode({'x': 1, 'y': 2, 'z': 3})


def test_serializers():
    serializers = json.Serializers()

    serializers.register(json.PrefixSerializer(
        target=str,
        prefix='upper',
        encode=lambda s: s.upper(),
        decode=lambda s: s.lower()
    ))

    serializers.register(json.DictionarySerializer(
        target=Point,
        keys=('x', 'y')
    ))

    assert serializers.encode('asdf') == '__upper__@ASDF'
    assert serializers.encode(Point(1, 2)) == {'x': 1, 'y': 2}
    assert serializers.decode('ASDF') == 'ASDF'
    assert serializers.decode('__upper__@ASDF') == 'asdf'
    assert serializers.decode({'x': 1, 'y': 2}) == Point(1, 2)
    assert serializers.decode({'x': 1, 'y': 2, 'z': 3}) == {
        'x': 1,
        'y': 2,
        'z': 3
    }


def test_serializable():
    serializers = json.Serializers()

    class SerializablePoint(Point, json.Serializable, keys=('x', 'y')):

        @classmethod
        def serializers(cls):
            return serializers  # for testing

    assert serializers.encode(SerializablePoint(1, 2)) == {'x': 1, 'y': 2}



# We use our own custom json implementation. In the libres library we made this
# configurable. Since onegov.core is a framework we don't do that though, we
# want all onegov.core applications with the same framework version to be able
# to read each others json.
#
# Therefore we use a common denominator kind of json encoder/decoder.
from sqlalchemy.ext.mutable import MutableDict
from sqlalchemy.types import TypeDecorator
from sqlalchemy.dialects.postgresql import JSONB


class JSON(TypeDecorator):
    """ A JSONB based type that coerces None's to empty dictionaries.

    That is, this JSONB column does not have NULL values, it only has
    falsy values (an empty dict).

    """

    impl = JSONB

    def process_bind_param(self, value, dialect):
        if value is None:
            return {}

        return value

    def process_result_value(self, value, dialect):
        if value is None:
            return {}

        return value


MutableDict.associate_with(JSON)



""" Provides a way to specify cronjobs which should be called at an exact time.

Example::

    @App.cronjob(hour=8, minute=0, timezone='Europe/Zurich')
    def cleanup_stuff(request):
        pass

Functions registered like this will be called through a regular request. That
means they are just like any other view and have all the power associated with
it.

WARNING: The function is called ONCE for EACH application id. So at the given
time there might be a number of calls more or less at the same time, run
against your function.

In terms of operations this means that there should be more processes than
application ids (tennants). In terms of development this means that you
should keep the time spent in a cronjob to an absolute minimum! Long running
jobs (more than 30 seconds) are reported.

This feature should really be used sparingly and only for things where the
exact time is relevant. Cleanup jobs for example should be combined with
delete/update/add jobs for example, not pushed off to a cronjob.

We also make sure that there's only one job each 5 minutes (at 00:00, 00:05,
and so on) and that there are no two jobs run at the same time.

For now cronjobs are furthermore limited to be run once a day. In the future
we might add a model more akin to classic cronjobs.

Finally note that cronjobs for any given application id are only run once the
first request to the application with that id has been made. The reason for
this is the fact that the background thread needs a real request to be
initialized.

"""

import pycurl

from contextlib import suppress
from datetime import datetime, timedelta
from onegov.core.framework import Framework, log
from onegov.core.utils import local_lock
from onegov.core.errors import AlreadyLockedError
from onegov.core.security import Public
from sedate import ensure_timezone, replace_timezone, utcnow, to_timezone
from threading import Thread
from time import sleep
from urllib.parse import quote_plus, unquote_plus


# the number of seconds the cronjob thread waits between polls, 45 was chosen
# because it means that each thread is run at least once a minute
CRONJOB_POLL_RESOLUTION = 45

# a job that takes longer than this (seconds) will be reported
CRONJOB_MAX_DURATION = 30


class Job(object):
    """ A single cron job. """

    def __init__(self, function, hour, minute, timezone):
        # the name is used to make sure the job is only run by one process
        # at a time in multi process environment. It needs to be unqiue
        # for each process but the same in all processes.
        self.name = function.__qualname__

        self.function = function
        self.hour = hour
        self.minute = minute
        self.timezone = ensure_timezone(timezone)

    @property
    def id(self):
        """ Internal id signed by the application. Used to access the job
        through an url which must be unguessable, but the same over many
        processes.

        The signature is only possible if ``self.app`` is present, which
        can only be set after the instance has already been created.

        See :meth:`as_request_call`.

        """

        assert self.app
        return quote_plus(self.app.sign(self.name))

    def as_request_call(self, request):
        """ Returns a new job which does the same as the old job, but it does
        so by calling an url which will execute the original job.

        """

        self.app = request.app
        url = request.link(self)

        def execute():
            # use curl for this because the urllib randomly hangs if we
            # use threads (it's fine with procsesses). requests has the same
            # problem because it uses urllib internally. My guess is there's
            # a race condition between code listening for new connections
            # and urllib. In any case, using pycurl gets rid of this problem
            # without introducing a rather expensive subprocess.
            c = pycurl.Curl()
            c.setopt(c.URL, url)
            c.setopt(pycurl.WRITEFUNCTION, lambda bytes: len(bytes))
            c.perform()
            c.close()

        return self.__class__(execute, self.hour, self.minute, self.timezone)

    def is_scheduled_at(self, point, timezone=None):
        """ Returns True if the job is scheduled to run at the given date.

        :point:
            The point in time. Needs to be a timezone aware datetime object,
            unless ``timezone`` is passed as well.

        :timezone:
            Optional. If the passed date is not timezone aware, the date is
            assumed to be from this timezone.

        :returns:
            True if the job is scheduled at this date. The maximum
            resolution is one minute. That means if the date given
            falls within the minute of the scheduled time it is assumed
            to be a match (i.e anywhere within 0' - 59'999.99999)

        """
        assert timezone or point.tzinfo

        point = point if point.tzinfo else replace_timezone(point, timezone)
        point = to_timezone(point, self.timezone)

        start = point.replace(
            hour=point.hour if self.hour == '*' else self.hour,
            minute=self.minute
        )

        return start <= point < (start + timedelta(minutes=1))


class ApplicationBoundCronjobs(Thread):
    """ A daemon thread which runs the cronjobs it is given in the background.

    The cronjobs are not actually run in the same thread. Instead a request to
    the actual cronjob is made. This way the thread is pretty much limited to
    basic IO work (GET request) and the actual cronjob is called with a normal
    request.

    Basically there is no difference between calling the url of the cronjob
    at a scheduled time and using this cronjob.

    This also avoids any and all locking problems as we don't have to write
    OneGov applications in a thread safe way. The actual work is always done
    on the main thread.

    Each application id is meant to have its own thread. To avoid extra
    work we make sure that only one thread per application exists on each
    server. This is accomplished by holding a local lock during
    the lifetime of the thread.

    WARNING This doesn't work on distributed systems. That is if the onegov
    processes are distributed over many servers the lock isn't shared.

    This can be achieved, but it's difficult to get right, so for now we
    do not implement it, as we do not have distributed systems yet.

    """

    def __init__(self, request, jobs):
        Thread.__init__(self, daemon=True)
        self.application_id = request.app.application_id
        self.jobs = tuple(job.as_request_call(request) for job in jobs)

        log.info(f"Starting Cronjob Thread for {self.application_id}")

        for job in jobs:
            log.info(f"Enabling Cronjob {job.name} ({request.link(job)})")

    def run(self):
        previous_check = datetime.min

        # the lock ensures that only one thread per application id is
        # in charge of running the scheduled jobs. If another thread already
        # has the lock, this thread will end immediately and be GC'd.
        with suppress(AlreadyLockedError):
            with local_lock('cronjobs-thread', self.application_id):
                while not sleep(CRONJOB_POLL_RESOLUTION):

                    # make sure we only run the jobs once a minute (we poll
                    # for it more often than that because we are not
                    # guaranteed to run in precise intervals)
                    this_check = datetime.utcnow()

                    if this_check.date() == previous_check.date():
                        if this_check.hour == previous_check.hour:
                            if this_check.minute == previous_check.minute:
                                continue

                    previous_check = this_check
                    self.run_scheduled_jobs()

        log.info("Stopping Cronjob Thread for {}, already locked!".format(
            self.application_id
        ))

    def run_scheduled_jobs(self):
        point = utcnow()

        for job in self.jobs:
            if job.is_scheduled_at(point):

                start = datetime.utcnow()
                job.function()
                duration = (datetime.utcnow() - start).total_seconds()

                if duration > CRONJOB_MAX_DURATION:
                    log.warning(
                        "The job for {} at {}:{} took too long ({}s)".format(
                            self.application_id, job.hour, job.minute, duration
                        )
                    )


@Framework.path(model=Job, path='/cronjobs/{id}')
def get_job(app, id):
    """ The internal path to the cronjob. The id can't be guessed. """
    name = app.unsign(unquote_plus(id))

    if name:
        return getattr(app.config.cronjob_registry, 'cronjobs', {}).get(name)


@Framework.view(model=Job, permission=Public)
def run_job(self, request):
    """ Executes the job. """
    self.function(request)


def register_cronjob(registry, function, hour, minute, timezone):
    assert minute % 5 == 0, "The cronjobs must be added in 5 minute increments"
    assert hour == '*' or isinstance(hour, int)

    if not hasattr(registry, 'cronjobs'):
        registry.cronjobs = {}
        registry.cronjob_threads = {}

    job = Job(function, hour, minute, timezone)
    registry.cronjobs[job.name] = job



import os.path
import pytest
import transaction
import yaml

from click.testing import CliRunner
from email.header import decode_header
from email.utils import parseaddr
from onegov.core.framework import Framework
from onegov.core.cli import cli, GroupContext
from smtplib import SMTPRecipientsRefused
from unittest.mock import patch


def test_group_context_without_schemas(postgres_dsn):
    ctx = GroupContext('/onegov_test/*', {
        'applications': [{
            'path': '/onegov_test/*',
            'application': 'onegov.core.Framework',
            'namespace': 'onegov_test',
            'configuration': {
                'dsn': postgres_dsn
            }
        }]
    })

    assert len(tuple(ctx.appcfgs)) == 1
    assert list(ctx.available_selectors) == ['/onegov_test/*']
    assert list(ctx.all_wildcard_selectors) == ['/onegov_test/*']
    assert list(ctx.all_specific_selectors) == []
    assert list(ctx.matches) == []


def test_group_context_with_schemas(postgres_dsn):
    class TestContext(GroupContext):

        def available_schemas(self, *args, **kwargs):
            return ['onegov_test-foo', 'onegov_test-bar']

    ctx = TestContext('/onegov_test/*', {
        'applications': [{
            'path': '/onegov_test/*',
            'application': 'onegov.core.Framework',
            'namespace': 'onegov_test',
            'configuration': {
                'dsn': postgres_dsn
            }
        }]
    })

    assert len(tuple(ctx.appcfgs)) == 1
    assert list(ctx.available_selectors) == [
        '/onegov_test/*',
        '/onegov_test/bar',
        '/onegov_test/foo'
    ]
    assert list(ctx.all_wildcard_selectors) == ['/onegov_test/*']
    assert list(ctx.all_specific_selectors) == [
        '/onegov_test/foo',
        '/onegov_test/bar'
    ]
    assert list(ctx.matches) == [
        '/onegov_test/foo',
        '/onegov_test/bar'
    ]

    ctx.selector = '/onegov_test/foo'
    assert list(ctx.matches) == [
        '/onegov_test/foo',
    ]

    ctx.selector = '/onegov_test/b??'
    assert list(ctx.matches) == [
        '/onegov_test/bar',
    ]


@pytest.mark.parametrize('cli', [{'singular': True}], indirect=True)
def test_create_command_group_single_path(cli, cli_config):
    runner = CliRunner()
    schemas = ['foobar-foo', 'foobar-bar']

    with patch.object(GroupContext, 'available_schemas', return_value=schemas):
        result = runner.invoke(cli, [
            '--config', cli_config, '--select', '/foobar/*', 'create'
        ])
        assert "The selector must match a single path" in result.output


@pytest.mark.parametrize('cli', [{'creates_path': True}], indirect=True)
def test_create_command_group_existing_path(cli, cli_config):
    runner = CliRunner()
    schemas = ['foobar-foo']

    with patch.object(GroupContext, 'available_schemas', return_value=schemas):
        result = runner.invoke(cli, [
            '--config', cli_config, '--select', '/foobar/foo', 'create'
        ])
        assert "may not reference an existing path" in result.output


@pytest.mark.parametrize('cli', [{'creates_path': True}], indirect=True)
def test_create_command_full_path(cli, cli_config):
    runner = CliRunner()
    schemas = ['foobar-foo']

    with patch.object(GroupContext, 'available_schemas', return_value=schemas):
        result = runner.invoke(cli, [
            '--config', cli_config, '--select', '/foobar', 'create'
        ])
        assert "must reference a full path" in result.output


@pytest.mark.parametrize('cli', [{'creates_path': True}], indirect=True)
def test_create_command_wildcard(cli, cli_config):

    runner = CliRunner()
    schemas = []

    with patch.object(GroupContext, 'available_schemas', return_value=schemas):
        result = runner.invoke(cli, [
            '--config', cli_config, '--select', '/foobar/*', 'create'
        ])
        assert "may not contain a wildcard" in result.output


@pytest.mark.parametrize('cli', [{'creates_path': True}], indirect=True)
def test_create_command_request_called(cli, cli_config):

    runner = CliRunner()
    schemas = []

    with patch.object(GroupContext, 'available_schemas', return_value=schemas):
        runner.invoke(cli, [
            '--config', cli_config, '--select', '/foobar/foo', 'create'
        ])

        assert cli.called
        assert cli.called_request


@pytest.mark.parametrize('cli', [{'default_selector': '*'}], indirect=True)
def test_create_command_default_selector(cli, cli_config):

    runner = CliRunner()
    schemas = ['foobar-foo']

    with patch.object(GroupContext, 'available_schemas', return_value=schemas):
        runner.invoke(cli, [
            '--config', cli_config, 'create'
        ])

        assert cli.called
        assert cli.called_request


@pytest.fixture(scope='function')
def maildir_app(temporary_directory):
    app_cfg = {
        'mail': {
            'marketing': {
                'sender': 'noreply@example.org',
                'use_directory': True,
                'directory': os.path.join(
                    temporary_directory, 'mails')
            },
            'transactional': {
                'sender': 'noreply@example.org',
                'use_directory': True,
                'directory': os.path.join(
                    temporary_directory, 'mails')
            }
        }
    }

    cfg = {
        'applications': [
            {
                'path': '/foobar/*',
                'application': 'onegov.core.Framework',
                'namespace': 'foobar',
                'configuration': app_cfg
            }
        ]
    }

    os.makedirs(os.path.join(temporary_directory, 'mails'))
    os.makedirs(os.path.join(temporary_directory, 'mails', 'new'))
    os.makedirs(os.path.join(temporary_directory, 'mails', 'cur'))
    os.makedirs(os.path.join(temporary_directory, 'mails', 'tmp'))

    with open(os.path.join(temporary_directory, 'onegov.yml'), 'w') as f:
        f.write(yaml.dump(cfg))

    app = Framework()
    app.configure_application(**app_cfg)

    return app


def test_sendmail(smtp, temporary_directory, maildir_app):
    app = maildir_app

    runner = CliRunner()
    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1]
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 0

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    transaction.commit()

    assert len(smtp.outbox) == 0

    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1]
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 1

    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1]
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 1

    message = smtp.outbox[0]

    assert message['Sender'] == 'Govikon <noreply@example.org>'
    assert message['From'] == 'Govikon <noreply@example.org>'
    assert message['Reply-To'] == 'Govikon <info@example.org>'
    assert message['Subject'] == 'Test E-Mail'
    assert message.get_payload()[0].as_string() == (
        'Content-Type: text/plain; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW1haWwgaXMganVzdCBhIHRlc3Q=\n'
    )
    assert message.get_payload()[1].as_string() == (
        'Content-Type: text/html; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        'VGhpcyBlLW1haWwgaXMganVzdCBhIHRlc3Q=\n'
    )


def test_sendmail_limit(smtp, temporary_directory, maildir_app):
    app = maildir_app

    runner = CliRunner()
    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1]
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 0

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    transaction.commit()

    assert len(smtp.outbox) == 0

    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1],
        '--limit', 1
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 1

    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1],
        '--limit', 2
    ])

    assert len(smtp.outbox) == 3

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    app.send_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mail",
        content="This e-mail is just a test"
    )

    transaction.commit()

    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1]
    ])

    assert len(smtp.outbox) == 5


def test_sendmail_unicode(smtp, temporary_directory, maildir_app):
    app = maildir_app

    app.send_email(
        reply_to='Gövikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mäil",
        content="👍"
    )

    transaction.commit()

    CliRunner().invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1]
    ])

    message = smtp.outbox[0]

    def decode(header):
        name, addr = parseaddr(header)

        try:
            name = decode_header(name)[0][0].decode('utf-8')
        except AttributeError:
            pass

        return name, addr

    assert decode(message['Sender']) == ("Gövikon", 'noreply@example.org')
    assert decode(message['From']) == ("Gövikon", 'noreply@example.org')
    assert decode(message['Reply-To']) == ("Gövikon", 'info@example.org')
    assert decode_header(message['Subject'])[0][0].decode('utf-8')\
        == "Test E-Mäil"

    assert message.get_payload()[0].as_string() == (
        'Content-Type: text/plain; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        '8J+RjQ==\n'
    )

    assert message.get_payload()[1].as_string() == (
        'Content-Type: text/html; charset="utf-8"\n'
        'MIME-Version: 1.0\n'
        'Content-Transfer-Encoding: base64\n\n'
        '8J+RjQ==\n'
    )


def test_sender_refused(smtp, temporary_directory, maildir_app):
    app = maildir_app
    app.send_email(
        reply_to='Gövikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Test E-Mäil",
        content="👍"
    )

    transaction.commit()

    def raise_error(*args, **kwargs):
        raise SMTPRecipientsRefused(recipients={'foo': 'bar'})

    with patch('smtplib.SMTP.sendmail', side_effect=raise_error):

        result = CliRunner().invoke(cli, [
            '--config', os.path.join(temporary_directory, 'onegov.yml'),
            'sendmail',
            '--hostname', smtp.address[0],
            '--port', smtp.address[1]
        ])

    assert len(smtp.outbox) == 0
    assert "Could not send e-mail: {'foo': 'bar'}" in result.output
    assert result.exit_code == 1


def test_sendmail_category(smtp, temporary_directory, maildir_app):
    app = maildir_app

    runner = CliRunner()
    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1]
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 0

    app.send_marketing_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Newsletter",
        content="Stuff that's happening"
    )

    app.send_transactional_email(
        reply_to='Govikon <info@example.org>',
        receivers=['recipient@example.org'],
        subject="Password Reset",
        content="Click here to reset"
    )

    transaction.commit()

    assert len(smtp.outbox) == 0

    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1],
        '--category', 'marketing'
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 1

    result = runner.invoke(cli, [
        '--config', os.path.join(temporary_directory, 'onegov.yml'),
        'sendmail',
        '--hostname', smtp.address[0],
        '--port', smtp.address[1],
        '--category', 'transactional'
    ])

    assert result.exit_code == 0
    assert len(smtp.outbox) == 2



""" Filestorage is a way to store files locally or on a remote server, with the
interface being the same, no matter where the files are stored.

Based on `<http://docs.pyfilesystem.org/en/latest/>`_

See :attr:`onegov.core.framework.Framework.filestorage` for more information.

"""

from fs.errors import IllegalBackReference
from onegov.core.framework import Framework
from onegov.core.crypto import random_token
from onegov.core.utils import render_file
from onegov.core.security import Public, Private


def random_filename():
    """ Returns a random filename that can't be guessed. """
    return random_token()


class FilestorageFile(object):
    """ Defines a static file served by the application. The difference
    between this and :class:`onegov.core.static.StaticFile` is the storage.

    Filestorage files are stored per application_id locally or on the cloud.
    Static files are the same for the whole application class and they are
    deployed statically. That means they are not content, but part of
    the distribution.

    Note that this is only used if the file is local. Files stored in the
    filestorage should be linked using
    :meth:`onegov.core.request.CoreRequest.filestorage_link`, which might
    result in a local path, for which this class is used. Or it might result in
    a remote path that is served by some different webserver.

    """
    storage = 'filestorage'

    def __init__(self, path):
        self.path = path

    @property
    def absorb(self):
        return self.path


@Framework.path(model=FilestorageFile, path='/files', absorb=True)
def get_filestorage_file(app, absorb):
    try:
        if app.filestorage.isfile(absorb):
            return FilestorageFile(absorb)
    except IllegalBackReference:
        return None


@Framework.view(model=FilestorageFile, render=render_file, permission=Public)
def view_filestorage_file(self, request):
    """ Renders the given filestorage file in the browser. """
    return getattr(request.app, self.storage).getsyspath(self.path)


@Framework.view(
    model=FilestorageFile, request_method='DELETE', permission=Private)
def delete_static_file(self, request):
    """ Deletes the given filestorage file. By default the permission is
    ``Private``. An application using the framework can override this though.

    Since a DELETE can only be sent through AJAX it is protected by the
    same-origin policy. That means that we don't need to use any CSRF
    protection here.

    That being said, browser bugs and future changes in the HTML standard
    make it possible for this to happen one day. Therefore, a time-limited
    token must be passed as query parameter to this function.

    New tokens can be acquired through ``request.new_csrf_token``.

    """
    request.assert_valid_csrf_token()
    request.app.filestorage.remove(self.path)



import morepath
import os.path

from onegov.core.framework import Framework
from onegov.core import utils
from webtest import TestApp as Client


def test_independence(temporary_directory):

    class App(Framework):
        pass

    app = App()
    app.configure_application(
        filestorage='fs.osfs.OSFS',
        filestorage_options={
            'root_path': temporary_directory
        }
    )
    app.namespace = 'tests'

    app.set_application_id('tests/foo')
    app.filestorage.settext('document.txt', 'foo')
    assert app.filestorage.getbytes('document.txt') == b'foo'
    assert app.filestorage.gettext('document.txt') == 'foo'

    app.set_application_id('tests/bar')
    assert not app.filestorage.exists('document.txt')
    app.filestorage.settext('document.txt', 'bar')
    assert app.filestorage.getbytes('document.txt') == b'bar'
    assert app.filestorage.gettext('document.txt') == 'bar'

    app.set_application_id('tests/foo')
    assert app.filestorage.getbytes('document.txt') == b'foo'
    assert app.filestorage.gettext('document.txt') == 'foo'

    assert os.path.isdir(os.path.join(temporary_directory, 'tests-foo'))
    assert os.path.isdir(os.path.join(temporary_directory, 'tests-bar'))
    assert os.path.isfile(
        os.path.join(temporary_directory, 'tests-foo/document.txt'))
    assert os.path.isfile(
        os.path.join(temporary_directory, 'tests-bar/document.txt'))


def test_filestorage(temporary_directory, redis_url):

    class App(Framework):
        pass

    @App.path('/')
    class Model(object):
        pass

    @App.path('/login')
    class Login(object):
        pass

    @App.view(model=Model)
    def view_file(self, request):
        return request.filestorage_link(request.params.get('file'))

    @App.view(model=Login)
    def view_login(self, request):

        @request.after
        def remember_login(response):
            request.app.remember_identity(response, request, morepath.Identity(
                userid=request.params.get('userid'),
                role='admin',
                application_id=request.app.application_id
            ))

    @App.view(model=Model, name='csrf-token')
    def view_csrf_token(self, request):
        return request.new_csrf_token()

    utils.scan_morepath_modules(App)
    App.commit()

    app = App()
    app.configure_application(
        filestorage='fs.osfs.OSFS',
        filestorage_options={
            'root_path': temporary_directory
        },
        identity_secure=False,
        redis_url=redis_url
    )
    app.namespace = 'tests'
    app.set_application_id('tests/foo')
    app.filestorage.settext('test.txt', 'asdf')
    app.filestorage.settext('readme', 'readme')

    client = Client(app)
    assert client.get('/?file=test.txt').text\
        == 'http://localhost/files/test.txt'
    assert client.get('/?file=asdf.txt').text == ''

    assert client.get('/files/test.txt').text == 'asdf'
    assert client.get('/files/test.txt').content_type == 'text/plain'

    assert client.get('/files/readme').text == 'readme'
    assert client.get('/files/readme').content_type == 'text/plain'

    app.set_application_id('tests/bar')

    client = Client(app)
    assert client.get('/?file=test.txt').text == ''
    assert client.get('/?file=asdf.txt').text == ''

    app.set_application_id('tests/foo')
    assert client.get('/files/readme').status_code == 200
    assert client.delete(
        '/files/readme', expect_errors=True).status_code == 403

    anonymous_csrf_token = client.get('/csrf-token').text.strip()
    client.get('/login?userid=test')
    logged_in_csrf_token = client.get('/csrf-token').text.strip()

    admin_client = Client(app)
    admin_client.get('/login?userid=admin')
    admin_csrf_token = admin_client.get('/csrf-token').text.strip()

    assert client.delete(
        '/files/readme', expect_errors=True).status_code == 403

    protected_url = '/files/readme?csrf-token={}'.format(anonymous_csrf_token)
    assert client.delete(protected_url, expect_errors=True).status_code == 403

    protected_url = '/files/readme?csrf-token={}'.format(admin_csrf_token)
    assert client.delete(protected_url, expect_errors=True).status_code == 403

    protected_url = '/files/readme?csrf-token={}'.format(logged_in_csrf_token)
    assert client.delete(protected_url, expect_errors=True).status_code == 200

    assert client.get('/files/readme', expect_errors=True).status_code == 404
    assert client.delete(
        '/files/readme', expect_errors=True).status_code == 404



from onegov.core.orm.types import JSON
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import deferred
from sqlalchemy.schema import Column


IMMUTABLE_TYPES = (int, float, complex, str, tuple, frozenset, bytes)


class ContentMixin(object):
    """ Mixin providing a meta/content JSON pair. Meta is a JSON column loaded
    with each request, content is a JSON column loaded deferred (to be shown
    only in the detail view).

    """

    #: metadata associated with the form, for storing small amounts of data
    @declared_attr
    def meta(cls):
        return Column(JSON, nullable=False, default=dict)

    #: content associated with the form, for storing things like long texts
    @declared_attr
    def content(cls):
        return deferred(Column(JSON, nullable=False, default=dict))


def is_valid_default(default):
    if default is None:
        return True

    if callable(default):
        return True

    if isinstance(default, IMMUTABLE_TYPES):
        return True

    return False


class dict_property(object):
    """ Enables access of dictionaries through properties.

    Usage::

        class Model(ContentMixin):
            access_times = dict_property('meta')

    This creates a property that accesses the meta directory with the key
    'access_times'. The key is implicitly copied from the definition.

    Another way of writing this out would be::

        class Model(ContentMixin):
            access_times = dict_property('meta', 'access_times')

    As is apparent, the 'access_times' key is duplicated in this case. Usually
    you do not need to provide the name. The exception being if you want
    the property name and the dictionary key to differ::

        class Model(ContentMixin):
            access_times = dict_property('meta', 'access')

    Here, the key in the dictionary is 'access', while the property is
    'access_times'.

    Since we often use the same kind of dictionaries we can use the builtin
    properties that are scoped to a specific dictionary::

        class Model(ContentMixin):
            access_times = meta_property()

    This is equivalent to the initial example.

    We can also create our own scoped properties as follows:

        foo_property = dict_property_factory('foo')

        class Model(object):

            foo = {}

            bar = foo_property()

    Here, model.bar would read model.foo['bar'].

    Dict properties are compatible with typical python properties, so the
    usual getter/setter/deleter methods are also available::

        class Model(ContentMixin):
            content = meta_property()

            @content.setter
            def set_content(self, value):
                self.meta['content'] = value
                self.meta['content_html'] = to_html(value)

            @content.deleter
            def del_content(self):
                del self.meta['content']
                del self.meta['content_html']

    """

    def __init__(self, attribute, key=None, default=None):
        assert is_valid_default(default)
        self.attribute = attribute
        self.key = key
        self.default = default

        self.custom_getter = None
        self.custom_setter = None
        self.custom_deleter = None

    def __set_name__(self, owner, name):
        """ Sets the dictionary key, if none is provided. """

        if self.key is None:
            self.key = name

    @property
    def getter(self):
        def wrapper(fn):
            self.custom_getter = fn
            return self

        return wrapper

    @property
    def setter(self):
        def wrapper(fn):
            self.custom_setter = fn
            return self

        return wrapper

    @property
    def deleter(self):
        def wrapper(fn):
            self.custom_deleter = fn
            return self

        return wrapper

    def __get__(self, instance, owner):

        # do not implement class-only behaviour
        if instance is None:
            return

        # pass control wholly to the custom getter if available
        if self.custom_getter:
            return self.custom_getter(instance)

        # get the value in the dictionary
        data = getattr(instance, self.attribute, None)

        if data is not None and self.key in data:
            return data[self.key]

        # fallback to the default
        return self.default() if callable(self.default) else self.default

    def __set__(self, instance, value):

        # create the dictionary if it does not exist yet
        if getattr(instance, self.attribute) is None:
            setattr(instance, self.attribute, {})

        # pass control to the custom setter if available
        if self.custom_setter:
            return self.custom_setter(instance, value)

        # fallback to just setting the value
        getattr(instance, self.attribute)[self.key] = value

    def __delete__(self, instance):

        # pass control to the custom deleter if available
        if self.custom_deleter:
            return self.custom_deleter(instance)

        # fallback to just removing the value
        del getattr(instance, self.attribute)[self.key]


def dict_property_factory(attribute):
    def factory(*args, **kwargs):
        return dict_property(attribute, *args, **kwargs)

    return factory


content_property = dict_property_factory('content')
data_property = dict_property_factory('data')
meta_property = dict_property_factory('meta')

# for backwards compatibility, might be removed in the future
dictionary_based_property_factory = dict_property_factory



""" Sigh. Here we go again, *another* json implementation with support for:

- date
- datetime
- time

Because nobody else does all of these. And if they do (like standardjson), they
don't support decoding...

"""

import datetime
import isodate
import json
import re
import types

from decimal import Decimal
from itertools import chain
from onegov.core.cache import lru_cache


class Serializer(object):
    """ Provides a way to encode all objects of a given class or its
    subclasses to and from json.

    """

    def __init__(self, target):
        assert isinstance(target, type), "expects a class"
        self.target = target

    def encode(self, obj):
        raise NotImplementedError

    def decode(self, value):
        raise NotImplementedError


class PrefixSerializer(Serializer):
    """ Serializes objects to a string with a prefix.

    Resulting json values take the form of __prefix__@<value>, where <value>
    is the encoded value and __prefix__@ is the prefix that is used to
    differentiate between normal strings and encoded strings.

    Note that the part after the prefix is user-supplied and possibly unsafe.
    So something like an 'eval' should be out of the question!

    """

    prefix_format = '__{}__@{}'
    prefix_expression = re.compile(r'__(?P<prefix>[a-zA-Z]+)__@')
    prefix_characters = re.compile(r'[a-zA-Z-]+')

    def __init__(self, target, prefix, encode, decode):
        super().__init__(target)

        assert self.prefix_characters.match(prefix)

        self.prefix = prefix
        self.prefix_length = len(self.prefix_format.format(prefix, ''))
        self._encode = encode
        self._decode = decode

    def encode(self, obj):
        return '__{}__@{}'.format(self.prefix, self._encode(obj))

    def decode(self, string):
        return self._decode(string[self.prefix_length:])


class DictionarySerializer(Serializer):
    """ Serialises objects that can be built with keyword arguments.

    For example::

        class Point(object):

            def __init__(self, x, y):
                self.x = x
                self.y = y

    Can be serialised using::

        DictionarySerializer(Point, ('x', 'y'))

    Which results in something like this in JSON::

        {'x': 1, 'y': 2}

    As the internal __dict__ represenation is of no concern, __slots__ may
    be used:

        class Point(object):

            __slots__ = ('x', 'y')

            def __init__(self, x, y):
                self.x = x
                self.y = y

    """

    def __init__(self, target, keys):
        super().__init__(target)

        self.keys = frozenset(keys)

    def encode(self, obj):
        return {k: getattr(obj, k) for k in self.keys}

    def decode(self, dictionary):
        return self.target(**dictionary)


class Serializers(object):
    """ Organises the different serializer implementations under a unifiying
    interface. This allows the actual encoder/decoder to call a single class
    without having to worry how the various serializers need to be looked up
    and called.

    """

    def __init__(self):
        self.by_prefix = {}
        self.by_keys = {}
        self.known_key_lengths = set()

    @property
    def registered(self):
        return chain(self.by_prefix.values(), self.by_keys.values())

    def register(self, serializer):
        if isinstance(serializer, PrefixSerializer):
            self.by_prefix[serializer.prefix] = serializer

        elif isinstance(serializer, DictionarySerializer):
            self.by_keys[serializer.keys] = serializer
            self.known_key_lengths.add(len(serializer.keys))

        else:
            raise NotImplementedError

    def serializer_for(self, value):
        if isinstance(value, str):
            return self.serializer_for_string(value)

        if isinstance(value, dict):
            return self.serializer_for_dict(value)

        if isinstance(value, type):
            return self.serializer_for_class(value)

    def serializer_for_string(self, string):
        match = PrefixSerializer.prefix_expression.match(string)
        return self.by_prefix.get(match.group('prefix') if match else None)

    def serializer_for_dict(self, dictionary):

        # we can exit early for all dictionaries which cannot possibly match
        # the keys we're looking for by comparing the number of keys in the
        # dictionary - this is much cheaper than the next lookup
        if len(dictionary.keys()) not in self.known_key_lengths:
            return

        return self.by_keys.get(frozenset(dictionary.keys()))

    @lru_cache(maxsize=16)
    def serializer_for_class(self, cls):
        matches = (s for s in self.registered if issubclass(cls, s.target))
        return next(matches, None)

    def encode(self, value):
        serializer = self.serializer_for(value.__class__)

        if serializer:
            return serializer.encode(value)

        if isinstance(value, types.GeneratorType):
            return tuple(v for v in value)

        raise TypeError('{} is not JSON serializable'.format(repr(value)))

    def decode(self, value):
        serializer = self.serializer_for(value)

        if serializer:
            value = serializer.decode(value)
        elif isinstance(value, dict):
            for k, v in value.items():
                value[k] = self.decode(v)

        return value


# The builtin serializers
default_serializers = Serializers()

default_serializers.register(PrefixSerializer(
    prefix='datetime',
    target=datetime.datetime,
    encode=isodate.datetime_isoformat,
    decode=isodate.parse_datetime
))

default_serializers.register(PrefixSerializer(
    prefix='time',
    target=datetime.time,
    encode=isodate.time_isoformat,
    decode=isodate.parse_time
))

default_serializers.register(PrefixSerializer(
    prefix='date',
    target=datetime.date,
    encode=isodate.date_isoformat,
    decode=isodate.parse_date
))

default_serializers.register(PrefixSerializer(
    prefix='decimal',
    target=Decimal,
    encode=str,
    decode=Decimal
))


class Serializable(object):
    """ Classes inheriting from this base are serialised using the
    :class:`DictionarySerializer` class.

    The keys that should be used need to be specified as follows::

        class Point(Serializable, keys=('x', 'y')):

            def __init__(self, x, y):
                self.x = x
                self.y = y

    """

    @classmethod
    def serializers(cls):
        return default_serializers  # for testing

    def __init_subclass__(cls, keys, **kwargs):
        super().__init_subclass__(**kwargs)

        cls.serialized_keys = keys
        cls.serializers().register(DictionarySerializer(
            target=cls,
            keys=keys
        ))


def dumps(obj, **extra):
    if obj is not None:
        return json.dumps(
            obj,
            default=default_serializers.encode,
            separators=(',', ':'),
            **extra)

    return None


def loads(txt, **extra):
    if txt is not None:
        return json.loads(
            txt,
            object_hook=default_serializers.decode,
            **extra)

    return {}


def dump(data, fp, *args, **kwargs):
    return fp.write(dumps(data, *args, **kwargs))


def load(fp, *args, **kwargs):
    return loads(fp.read(), *args, **kwargs)



from sqlalchemy.dialects.postgresql import HSTORE as HSTOREBase
from sqlalchemy.ext.mutable import MutableDict


class HSTORE(HSTOREBase):
    """ Extends the default HSTORE type to make it mutable by default. """


MutableDict.associate_with(HSTORE)



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""
from onegov.core.upgrade import upgrade_task
from onegov.core.orm import Base, find_models
from onegov.core.orm.abstract import Associable
from onegov.core.orm.types import JSON
from sqlalchemy import inspect, text
from sqlalchemy.exc import NoInspectionAvailable


@upgrade_task('Drop primary key from associated tables')
def drop_primary_key_from_associated_tables(context):
    bases = set()

    for cls in find_models(Base, lambda cls: issubclass(cls, Associable)):
        bases.add(cls.association_base())

    for base in bases:
        for link in base.registered_links.values():
            if context.has_table(link.table.name):

                # XXX do not use this kind of statement outside upgrades!
                command = """
                    ALTER TABLE {table} DROP CONSTRAINT IF EXISTS {table}_pkey
                """.format(table=link.table.name)

                context.operations.execute(command)


@upgrade_task('Migrate to JSONB', always_run=True, raw=True)
def migrate_to_jsonb(connection, schemas):
    """ Migrates all text base json columns to jsonb. """

    def json_columns(cls):
        try:
            for column in inspect(cls).columns:
                if isinstance(column.type, JSON):
                    yield column
        except NoInspectionAvailable:
            pass

    classes = list(find_models(Base, is_match=lambda cls: True))

    # XXX onegov.libres (but not libres itself) uses json with a different orm
    # base - so we need to included it manually
    try:
        from libres.db.models import ORMBase
        classes.extend(find_models(ORMBase, is_match=lambda cls: True))
    except ImportError:
        pass

    columns = list(c for cls in classes for c in json_columns(cls))

    if not columns:
        return False

    text_columns = list(
        r.identity for r in connection.execute(text("""
            SELECT concat_ws(':', table_schema, table_name, column_name)
                AS identity
              FROM information_schema.columns
             WHERE data_type != 'jsonb'
               AND table_schema IN :schemas
               AND column_name IN :names
        """), schemas=tuple(schemas), names=tuple(c.name for c in columns))
    )

    for schema in schemas:
        for column in columns:
            identity = ':'.join((schema, column.table.name, column.name))

            if identity not in text_columns:
                continue

            # XXX do not use this kind of statement outside upgrades!
            connection.execute("""
                ALTER TABLE "{schema}".{table}
                ALTER COLUMN {column}
                TYPE JSONB USING {column}::jsonb
            """.format(
                schema=schema,
                table=column.table.name,
                column=column.name
            ))

            # commits/rolls back the current transaction (to keep the number
            # of required locks to a minimum)
            yield True



""" Integrates the Chameleon template language.

This is basically a copy of more.chameleon, with the additional inclusion of a
gettext translation function defined by :mod:`onegov.core.i18n`.

To use a chameleon template, applications have to specify the templates
directiory, in addition to inheriting from
:class:`onegov.core.framework.Framework`.

For example::

    from onegov.core.framework import Framework

    class App(Framework):
        pass

    @App.template_directory()
    def get_template_directory():
        return 'templates'

    @App.path()
    class Root(object):
        pass

    @App.html(model=Root, template='index.pt')
    def view_root(self, request):
        return {
            'title': 'The Title'
        }

The folder can either be a directory relative to the app class or an absolute
path.

"""

import os.path

from cached_property import cached_property
from chameleon import PageTemplateLoader, PageTemplateFile
from chameleon.tal import RepeatDict
from chameleon.utils import Scope, decode_string
from onegov.core.framework import Framework


AUTO_RELOAD = os.environ.get('ONEGOV_DEVELOPMENT') == '1'


def get_default_vars(request, content, suppress_global_variables=False):

    default = {
        'request': request,
        'translate': request.get_translate(for_chameleon=True)
    }

    default.update(content)

    if suppress_global_variables:
        return default
    else:
        return request.app.config.templatevariables_registry.get_variables(
            request, default)


class TemplateLoader(PageTemplateLoader):
    """ Extends the default page template loader with the ability to
    lookup macros in various folders.

    """

    @cached_property
    def macros(self):
        return MacrosLookup(self.search_path, name='macros.pt')

    @cached_property
    def mail_macros(self):
        return MacrosLookup(self.search_path, name='mail_macros.pt')


class MacrosLookup(object):
    """ Takes a list of search paths and provides a lookup for macros.

    This means that when a macro is access through this lookup, it will travel
    up the search path of the template loader, to look for the macro and
    return with the first match.

    As a result, it is possible to have a macros.pt file in each search path
    and have them act as if they were one file, with macros further up the
    list of paths having precedence over the macros further down the path.

    For example, given the search paths 'foo' and 'bar', foo/macros.pt could
    define 'users' and 'page', while bar/macros.pt could define 'users' and
    'site'. In the lookup this would result in 'users' and 'page' being loaded
    loaded from foo and 'site' being loaded from bar.

    """

    def __init__(self, search_paths, name='macros.pt'):
        paths = (os.path.join(base, name) for base in search_paths)
        paths = (path for path in paths if os.path.isfile(path))

        # map each macro name to a template
        self.lookup = {
            name: template
            for template in (
                PageTemplateFile(path, search_paths, auto_reload=AUTO_RELOAD)
                for path in reversed(list(paths))
            )
            for name in template.macros.names
        }

    def __getitem__(self, name):
        # macro names in chameleon are normalized internally and we need
        # to do the same to get the correct name in any case:
        name = name.replace('-', '_')
        return self.lookup[name].macros[name]


@Framework.template_loader(extension='.pt')
def get_template_loader(template_directories, settings):
    """ Returns the Chameleon template loader for templates with the extension
    ``.pt``.

    """

    return TemplateLoader(
        template_directories,
        default_extension='.pt',
        prepend_relative_search_path=False,
        auto_reload=AUTO_RELOAD,
    )


@Framework.template_render(extension='.pt')
def get_chameleon_render(loader, name, original_render):
    """ Returns the Chameleon template renderer for the required template.

    """
    template = loader.load(name, 'xml')

    def render(content, request):

        variables = get_default_vars(request, content)
        return original_render(template.render(**variables), request)

    return render


def render_template(template, request, content,
                    suppress_global_variables='infer'):
    """ Renders the given template. Use this if you need to get the rendered
    value directly. If oyu render a view, this is not needed!

    By default, mail templates (templates strting with 'mail_') skip the
    inclusion of global variables defined through the template_variables
    directive.

    """

    if suppress_global_variables == 'infer':
        suppress_global_variables = template.startswith('mail_')

    registry = request.app.config.template_engine_registry
    template = registry._template_loaders['.pt'][template]

    variables = get_default_vars(
        request, content, suppress_global_variables=suppress_global_variables)

    return template.render(**variables)


def render_macro(macro, request, content, suppress_global_variables=True):
    """ Renders a :class:`chameleon.zpt.template.Macro` like this::

        layout.render_macro(layout.macros['my_macro'], **vars)

    This code is basically a stripped down version of this:
    `<https://github.com/malthe/chameleon/blob\
    /257c9192fea4b158215ecc4f84e1249d4b088753/src/chameleon\
    /zpt/template.py#L206>`_.

    As such it doesn't treat chameleon like a black box and it will probably
    fail one day in the future, if Chameleon is refactored. Our tests will
    detect that though.

    """

    if not hasattr(request, '_macro_variables'):
        variables = get_default_vars(
            request=request,
            content={},
            suppress_global_variables=suppress_global_variables
        )

        variables.setdefault('__translate', variables['translate'])
        variables.setdefault('__convert', variables['translate'])
        variables.setdefault('__decode', decode_string)
        variables.setdefault('target_language', None)
        request._macro_variables = variables
    else:
        variables = request._macro_variables.copy()

    variables.update(content)
    variables['repeat'] = RepeatDict({})

    stream = list()
    macro.include(stream, Scope(variables), {})

    return ''.join(stream)



""" Simple redirects for renamed paths using a generic redirect model.

For static paths::

    @App.path('/old-path')
    class OldPathRedirect(Redirect):
        to = '/new-path'

For wildcard paths (e.g. /old-pages/my-page to /new-pages/my-page)::

    @App.path('/old-path', absorb=True)
    class OldPagesRedirect(Redirect):
        to = '/new-pages

"""

from onegov.core import Framework
from onegov.core.security import Public


class Redirect(object):
    to = None

    def __init__(self, absorb=None):
        assert self.to and not self.to.endswith('/')

        if absorb:
            self.to = self.to + '/' + absorb


@Framework.view(model=Redirect, permission=Public)
def view_redirect(self, request):
    return request.redirect(self.to)



import inspect
import importlib
import pkg_resources
import transaction

from alembic.migration import MigrationContext
from alembic.operations import Operations
from contextlib import contextmanager
from inspect import getmembers, isfunction, ismethod
from onegov.core.orm import Base, find_models
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import JSON
from sqlalchemy import Column, Text
from sqlalchemy import create_engine
from sqlalchemy.engine.reflection import Inspector
from sqlalchemy.pool import StaticPool
from sqlalchemy.orm import load_only
from toposort import toposort, toposort_flatten


class UpgradeState(Base, TimestampMixin):
    """ Keeps the state of all upgrade steps over all modules. """

    __tablename__ = 'upgrades'

    # the name of the module (e.g. onegov.core)
    module = Column(Text, primary_key=True)

    # a json holding the state of the upgrades
    state = Column(JSON, nullable=False, default=dict)

    @property
    def executed_tasks(self):
        if not self.state:
            return set()
        else:
            return set(self.state.get('executed_tasks', []))

    def was_already_executed(self, task):
        return task.task_name in self.executed_tasks

    def mark_as_executed(self, task):
        if not self.state:
            self.state = {}

        if 'executed_tasks' in self.state:
            self.state['executed_tasks'].append(task.task_name)
            self.state['executed_tasks'] = list(
                set(self.state['executed_tasks']))
        else:
            self.state['executed_tasks'] = [task.task_name]

        self.state.changed()


def get_distributions_with_entry_map(key):
    """ Iterates through all distributions with entry_maps and yields
    each distribution along side the entry map with the given key.

    """
    for distribution in pkg_resources.working_set:
        if hasattr(distribution, 'get_entry_map'):
            entry_map = distribution.get_entry_map(key)

            if entry_map:
                yield distribution, entry_map


def get_upgrade_modules():
    """ Returns all modules that registered themselves for onegov.core
    upgrades like this::

        entry_points={
            'onegov.core': [
                'upgrade = onegov.mypackage.upgrade'
            ]
        }

    """
    yield 'onegov.core', importlib.import_module('onegov.core.upgrades')

    for distribution, entry_map in get_distributions_with_entry_map('onegov'):
        if 'upgrade' in entry_map:
            yield (
                '.'.join(entry_map['upgrade'].module_name.split('.')[:2]),
                importlib.import_module(entry_map['upgrade'].module_name)
            )


class upgrade_task(object):
    """ Marks the decorated function as an upgrade task. Upgrade tasks should
    be defined outside classes (except for testing) - that is in the root of
    the module (directly in onegov/form/upgrades.py for example).

    Each upgrade task has the following properties:

    :name:
        The name of the task in readable English. Do not change the name of
        the task once you pushed your changes. This name is used as an id!

    :always_run:
        By default, tasks are run once on if they haven't been run yet, or if
        the module was not yet under upgrade control (it just added itself
        through the entry_points mechanism).

        There are times where this is not wanted.

        If you enable upgrades in your module and you plan to run an upgrade
        task in the same release, then you need to 'always_run' and detect
        yourself if you need to do any changes.

        If you need to enable upgrades and run a task at the same time which
        can only be run once and is impossible to make idempotent, then you
        need to make two releases. One that enables upgade control and a second
        one that

        Also, if you want to run an upgrade task every time you upgrade (for
        example, for maintenance), then you should also use it.

        The easiest way to use upgrade control is to enable it from the first
        release on.

        Another way to tackle this is to write update methods idempotent as
        often as possible and to always run them.

    :requires:
        A reference to another projects upgrade task that should be run first.

        For example::

            requires='onegov.form:Add new form field'

        If the reference cannot be found, the upgrade fails.

    :raw:
        A raw task is run without setting up a request. Instead a database
        connection and a list of targeted schemas is passed. It's up to the
        upgrade function to make sure the right schemas are targeted through
        that connection!

        This is useful for upgrades which need to execute some raw SQL that
        cannot be executed in a request context. For example, this is a way
        to migrate the upgrade states table of the upgrade itself.

        Raw tasks are always run and there state is not recorded. So the
        raw task needs to be idempotent and return False if there was no
        change.

        Only use this if you have no other way. This is a very blunt instrument
        only used under special circumstances.

        Note, raw tasks may be normal function OR generators. If they are
        generators, then with each yield the transaction is either commited
        or rolled-back depending on the dry-run cli argument.

    Always run tasks may return ``False`` if they wish to be considered
    as not run (and therefore not shown in the upgrade runner).

    """

    def __init__(self, name, always_run=False, requires=None, raw=False):
        if raw:
            assert always_run, "raw tasks must always run"
            assert not requires, "raw tasks may not require other tasks"

        self.name = name
        self.always_run = always_run
        self.requires = requires
        self.raw = raw

    def __call__(self, fn):
        fn.task_name = self.name
        fn.always_run = self.always_run
        fn.requires = self.requires
        fn.raw = self.raw
        return fn


def is_task(function):
    """ Returns True if the given function is an uprade task. """
    if not (isfunction(function) or ismethod(function)):
        return False

    return hasattr(function, 'task_name')


def get_module_tasks(module):
    """ Goes through a module or class and returns all upgrade tasks. """
    for name, function in getmembers(module, is_task):
        yield function


def get_tasks_by_id(upgrade_modules=None):
    """ Takes a list of upgrade modules or classes and returns the tasks
    keyed by id.

    """

    tasks = {}
    fn_names = set()

    for distribution, upgrade_module in upgrade_modules:

        for function in get_module_tasks(upgrade_module):
            task_id = ':'.join((distribution, function.task_name))

            assert task_id not in tasks, "Duplicate task"
            tasks[task_id] = function

            # make sure we don't have duplicate function names - it works, but
            # it makes debugging harder
            assert function.__name__ not in fn_names, "Duplicate function name"
            fn_names.add(function.__name__)

    return tasks


def get_module_order_key(tasks):
    """ Returns a sort order key which orders task_ids in order of their
    module dependencies. That is a task from onegov.core is sorted before
    a task in onegov.user, because onegov.user depends on onegov.core.

    This is used to order unrelated tasks in a sane way.

    """
    modules = set()

    for task in tasks:
        modules.add(task.split(':')[0])

    graph = {}
    packages = pkg_resources.working_set

    for module in modules:
        if module not in graph:
            graph[module] = set()

        if module in packages.by_key:
            for dependency in packages.by_key[module].requires():
                graph[module].add(dependency.key)

    sorted_modules = {
        task_id: ix for ix, task_id in enumerate(toposort_flatten(graph))
    }

    def sortkey(task_id):
        module, name = task_id.split(':', 1)
        return (
            # sort by the topological order
            sorted_modules.get(module, float('inf')),

            # then by module
            module,

            # then by appearance of update function in the code
            tasks[task_id].__code__.co_firstlineno
        )

    return sortkey


def get_tasks(upgrade_modules=None):
    """ Takes a list of upgrade modules or classes and returns the
    tasks that should be run in the order they should be run.

    The order takes dependencies into account.

    """

    tasks = get_tasks_by_id(upgrade_modules or get_upgrade_modules())

    for task_id, task in tasks.items():
        if task.requires:
            assert not tasks[task.requires].raw, "Raw tasks cannot be required"

    graph = {}

    for task_id, task in tasks.items():
        if task_id not in graph:
            graph[task_id] = set()

        if task.requires is not None:
            graph[task_id].add(task.requires)

    sorted_tasks = []
    order_key = get_module_order_key(tasks)

    for task_ids in toposort(graph):
        sorted_tasks.extend(sorted(task_ids, key=order_key))

    return [(task_id, tasks[task_id]) for task_id in sorted_tasks]


def register_modules(session, modules, tasks):
    """ Sets up the state tracking for all modules. Initially, all tasks
    are marekd as executed, because we assume tasks to upgrade older
    deployments to a new deployment.

    If this is a new deployment we do not need to execute these tasks.

    Tasks where this is not desired should be marked as 'always_run'.
    They will then manage their own state (i.e. check if they need to
    run or not).

    This function is idempotent.

    """

    for module, upgrade_module in modules:
        query = session.query(UpgradeState)
        query = query.filter(UpgradeState.module == module)

        if query.first():
            continue

        session.add(
            UpgradeState(module=module, state={
                'executed_tasks': [
                    task.task_name for task_id, task in tasks
                    if task_id.startswith(module)
                ]
            })
        )

    session.flush()


def register_all_modules_and_tasks(session):
    """ Registers all the modules and all the tasks. """
    register_modules(session, get_upgrade_modules(), get_tasks())


class UpgradeTransaction(object):
    """ Holds the session and the alembic operations connection together and
    commits/aborts both at the same time.

    Not really a two-phase commit solution. That could be accomplished but
    for now this suffices.

    """

    def __init__(self, context):
        self.operations_transaction = context.operations_connection.begin()
        self.session = context.session

    def flush(self):
        self.session.flush()

    def commit(self):
        self.operations_transaction.commit()
        transaction.commit()

    def abort(self):
        self.operations_transaction.rollback()
        transaction.abort()


class UpgradeContext(object):
    """ Holdes the context of the upgrade. An instance of this is passed
    to each upgrade task.

    """

    def __init__(self, request):

        # we need to reset the request session property every time because
        # we reuse the request over various session managers
        if 'session' in request.__dict__:
            del request.__dict__['session']

        self.request = request
        self.session = request.session
        self.app = request.app
        self.schema = request.app.session_manager.current_schema
        self.engine = self.session.bind

        # The locale of the upgrade is the default locale
        request.app.session_manager.set_locale(
            current_locale=self.request.locale,
            default_locale=self.request.default_locale
        )

        # Make sure the connection is the same for the session, the engine
        # and the alembic migrations context. Otherwise the upgrade locks up.
        # If that happens again, test with this:
        #
        #       http://docs.sqlalchemy.org/en/latest/core/pooling.html
        #       #sqlalchemy.pool.AssertionPool
        #
        self.operations_connection = self.session._connection_for_bind(
            self.session.bind)
        self.operations = Operations(
            MigrationContext.configure(self.operations_connection))

    def begin(self):
        return UpgradeTransaction(self)

    def has_column(self, table, column):
        inspector = Inspector(self.operations_connection)
        return column in {c['name'] for c in inspector.get_columns(
            table, schema=self.schema
        )}

    def has_table(self, table):
        inspector = Inspector(self.operations_connection)
        return table in inspector.get_table_names(schema=self.schema)

    def models(self, table):
        def has_matching_tablename(model):
            if not hasattr(model, '__tablename__'):
                return False  # abstract bases

            return model.__tablename__ == table

        for base in self.request.app.session_manager.bases:
            yield from find_models(base, has_matching_tablename)

    def records_per_table(self, table, columns):
        for model in self.models(table):
            yield from self.session.query(model).options(
                load_only(*(c.name for c in columns))
            )

    @contextmanager
    def stop_search_updates(self):
        # XXX this would be better handled with a more general approach
        # that doesn't require knowledge of onegov.search
        getattr(self.app, 'es_orm_events', object()).stopped = True
        yield
        getattr(self.app, 'es_orm_events', object()).stopped = False

    def add_column_with_defaults(self, table, column, default):
        # XXX while adding default values we shouldn't reindex the data
        # since this is what the default add_column code does and will be
        # doing in the future when Postgres 11 supports defaults for
        # new columns - the way this is implemented now is with tight coupling
        # that we shouldn't do, but we want to replace this when Postgres 11
        # comes around
        with self.stop_search_updates():
            nullable_column = column.copy()
            nullable_column.nullable = True

            self.operations.add_column(table, nullable_column)

            # fill it with defaults
            for record in self.records_per_table(table, (column, )):
                value = default(record) if callable(default) else default
                setattr(record, column.name, value)

                if hasattr(getattr(record, column.name), 'changed'):
                    getattr(record, column.name).changed()

            # make sure all generated values are flushed
            self.session.flush()

            # activate non-null constraint
            if not column.nullable:
                self.operations.alter_column(
                    table, column.name, nullable=False)

            # propagate to db
            self.session.flush()


class RawUpgradeRunner(object):
    """ Runs the given raw tasks. """

    def __init__(self, tasks, commit=True,
                 on_task_success=None, on_task_fail=None):
        self.tasks = tasks
        self.commit = commit

        self._on_task_success = on_task_success
        self._on_task_fail = on_task_fail

    def run_upgrade(self, dsn, schemas):

        # it's possible for applications to exist without schemas, if the
        # application doesn't use multi-tennants and has not been run yet
        if not schemas:
            return 0

        engine = create_engine(dsn, poolclass=StaticPool)
        connection = engine.connect()
        executions = 0

        for id, task in self.tasks:
            t = connection.begin()

            try:
                success = False

                if inspect.isgeneratorfunction(task):
                    for changed in task(connection, schemas):
                        success = success or changed

                        if self.commit:
                            t.commit()
                        else:
                            t.rollback()

                        t = connection.begin()
                else:
                    success = task(connection, schemas)

                if success:
                    executions += 1
                    self._on_task_success(task)

            except Exception:
                t.rollback()
                self._on_task_fail(task)

                raise
            else:
                if self.commit:
                    t.commit()
                else:
                    t.rollback()

        return executions


class UpgradeRunner(object):
    """ Runs the given basic tasks. """

    def __init__(self, modules, tasks, commit=True,
                 on_task_success=None, on_task_fail=None):
        self.modules = modules
        self.tasks = tasks
        self.commit = commit
        self.states = {}

        self._on_task_success = on_task_success
        self._on_task_fail = on_task_fail

    def on_task_success(self, task):
        if self._on_task_success is not None:
            self._on_task_success(task)

    def on_task_fail(self, task):
        if self._on_task_fail is not None:
            self._on_task_fail(task)

    def get_state(self, context, module):
        if module not in self.states:
            query = context.session.query(UpgradeState)
            query = query.filter(UpgradeState.module == module)

            # assume that all modules have been ran through register_module
            self.states[module] = query.one()

        return context.session.merge(self.states[module], load=False)

    def register_modules(self, request):
        register_modules(request.session, self.modules, self.tasks)

        if self.commit:
            transaction.commit()

    def get_module_from_task_id(self, task_id):
        return task_id.split(':')[0]

    def run_upgrade(self, request):
        self.register_modules(request)

        tasks = ((self.get_module_from_task_id(i), t) for i, t in self.tasks)
        executed = 0

        for module, task in tasks:
            context = UpgradeContext(request)
            state = self.get_state(context, module)

            if not task.always_run and state.was_already_executed(task):
                continue

            upgrade = context.begin()

            try:
                result = task(context)

                # mark all tasks as executed, even 'always run' ones
                state.mark_as_executed(task)
                upgrade.flush()

                # always-run tasks which return False are considered
                # to have not run
                hidden = task.always_run and result is False

                if not hidden:
                    executed += 1
                    self.on_task_success(task)

            except Exception:
                upgrade.abort()
                self.on_task_fail(task)

                raise
            else:
                if self.commit:
                    upgrade.commit()
                else:
                    upgrade.abort()
            finally:
                context.session.invalidate()
                context.engine.dispose()

        return executed



import pytest

from onegov.core.html import sanitize_html, sanitize_svg


def test_sanitize_html():
    # this is really bleach's job, but we want to run the codepath anyway
    assert sanitize_html('') == ''
    assert sanitize_html('<script>') == '&lt;script&gt;'
    assert sanitize_html('<b foo="x" foo:bar="y">') == '<b></b>'


def test_sanitize_svg():
    with pytest.raises(AssertionError):
        assert sanitize_svg('<a xlink:href=”javascript:alert(9)”>')

    with pytest.raises(AssertionError):
        assert sanitize_svg('<a><![CDATA[ alert(); ]]></a>')

    with pytest.raises(AssertionError):
        assert sanitize_svg('<script>')

    with pytest.raises(AssertionError):
        assert sanitize_svg("<meta http-equiv='Set-Cookie' content='a=b' />")



import click
import email
import mailbox
import os
import platform
import subprocess
import sys

from cached_property import cached_property
from mailthon.middleware import TLS, Auth
from onegov.core.cli.core import command_group, pass_group_context
from onegov.core.mail import Postman
from onegov.core.upgrade import get_tasks
from onegov.core.upgrade import get_upgrade_modules
from onegov.core.upgrade import RawUpgradeRunner
from onegov.core.upgrade import UpgradeRunner
from onegov.server.config import Config
from smtplib import SMTPRecipientsRefused
from uuid import uuid4


#: onegov.core's own command group
cli = command_group()


@cli.command(context_settings={
    'matches_required': False,
    'default_selector': '*'
})
@click.option('--hostname', help="The smtp host")
@click.option('--port', help="The smtp port")
@click.option('--force-tls', default=False, is_flag=True,
              help="Force a TLS connection")
@click.option('--username', help="The username to authenticate", default=None)
@click.option('--password', help="The password to authenticate", default=None)
@click.option('--limit', default=25,
              help="Max number of mails to send before exiting")
@click.option('--category', help="Only send e-mails of the given category",
              default=None)
@pass_group_context
def sendmail(group_context,
             hostname, port, force_tls, username, password, limit, category):
    """ Iterates over all applications and processes the maildir for each
    application that uses maildir e-mail delivery.

    """
    exit_code = 0

    class Mail(object):

        def __init__(self, maildir, filename):
            self.maildir = maildir
            self.filename = filename

        @cached_property
        def message(self):
            return email.message_from_string(self.text)

        @cached_property
        def text(self):
            return self.maildir.get_file(self.filename).read().decode('utf-8')

        @property
        def category(self):
            return self.message.get('X-Category')

        def remove(self):
            self.maildir.remove(self.filename)

        def send(self, connection):
            connection.sendmail(
                self.message['From'],
                self.message['To'],
                self.text)

            return connection.noop()[0]

    # applications with a maildir configuration
    cfgs = (c for c in group_context.appcfgs if 'mail' in c.configuration)
    cfgs = (v for c in cfgs for v in c.configuration['mail'].values())
    cfgs = (c for c in cfgs if c.get('use_directory'))

    # non-empty maildirs
    dirs = set(c['directory'] for c in cfgs)
    dirs = (mailbox.Maildir(d, create=False) for d in dirs)
    dirs = (d for d in dirs if len(d))

    # emails
    mails = (Mail(d, f) for d in dirs for f in d.keys())
    mails = (m for m in mails if category is None or m.category == category)

    # load all the mails we're going to process
    if limit:
        mails = tuple(x for _, x in zip(range(limit), mails))
    else:
        mails = tuple(mails)

    if not mails:
        sys.exit(exit_code)

    # create the connection handler
    postman = Postman(hostname, port)

    if force_tls:
        postman.middlewares.append(TLS(force=True))

    if username:
        postman.middlewares.append(Auth(username, password))

    # send the messages
    with postman.connection() as connection:

        for mail in mails:
            try:
                status = mail.send(connection)
            except SMTPRecipientsRefused as e:
                exit_code = 1
                print(f"Could not send e-mail: {e.recipients}")
                mail.remove()
            else:
                if status == 250:
                    mail.remove()

    sys.exit(exit_code)


@cli.command()
@click.argument('server')
@click.argument('remote-config')
@click.option('--confirm/--no-confirm', default=True,
              help="Ask for confirmation (disabling this is dangerous!)")
@click.option('--no-filestorage', default=False, is_flag=True,
              help="Do not transfer the files")
@click.option('--no-database', default=False, is_flag=True,
              help="Do not transfer the database")
@pass_group_context
def transfer(group_context,
             server, remote_config, confirm, no_filestorage, no_database):
    """ Transfers the database and all files from a server running a
    onegov-cloud application and installs them locally, overwriting the
    local data!

    This command expects to have access to the remote server via ssh
    (no password) and to be run sudo without password. If this is too scary
    for you, feel free to write something saner.

    Only namespaces which are present locally and remotely are considered.

    So if you have a 'cities' namespace locally and a 'towns' namespace on
    the remote, nothing will happen.

    WARNING: This may delete local content!

    """

    if confirm:
        click.confirm(
            "Do you really want override all your local data?",
            default=False, abort=True
        )

    print("Parsing the remote application configuration")

    remote_dir = os.path.dirname(remote_config)
    remote_config = Config.from_yaml_string(
        subprocess.check_output([
            "ssh", server, "-C", "sudo cat '{}'".format(remote_config)
        ])
    )

    remote_applications = {a.namespace: a for a in remote_config.applications}

    # storages may be shared between applications, so we need to keep track
    fetched = set()

    def download_folder(remote, local):
        tar_filename = '/tmp/{}.tar.gz'.format(uuid4().hex)

        subprocess.check_output([
            'ssh', server, '-C', "sudo tar czvf '{}' -C '{}' .".format(
                tar_filename, remote
            )
        ])

        subprocess.check_output([
            'scp',
            '{}:{}'.format(server, tar_filename),
            '{}/transfer.tar.gz'.format(local)
        ])

        subprocess.check_output([
            'tar', 'xzvf', '{}/transfer.tar.gz'.format(local),
            '-C', local
        ])

        subprocess.check_output([
            'ssh', server, '-C', "sudo rm '{}'".format(tar_filename)
        ])

        subprocess.check_output([
            'rm', '{}/transfer.tar.gz'.format(local)
        ])

    # filter out namespaces on demand
    for appcfg in group_context.appcfgs:

        if appcfg.configuration.get('disable_transfer'):
            print("Skipping {}, transfer disabled".format(appcfg.namespace))
            continue

        if appcfg.namespace not in remote_applications:
            print("Skipping {}, not found on remote".format(appcfg.namespace))
            continue

        remotecfg = remote_applications[appcfg.namespace]

        if not no_filestorage:
            if remotecfg.configuration.get('filestorage').endswith('OSFS'):
                assert appcfg.configuration.get('filestorage').endswith('OSFS')

                print("Fetching remote filestorage")

                local_fs = appcfg.configuration['filestorage_options']
                remote_fs = remotecfg.configuration['filestorage_options']

                remote_storage = os.path.join(
                    remote_dir, remote_fs['root_path'])
                local_storage = os.path.join('.', local_fs['root_path'])

                if ':'.join((remote_storage, local_storage)) in fetched:
                    continue

                download_folder(remote_storage, local_storage)

                fetched.add(':'.join((remote_storage, local_storage)))

            remote_backend = remotecfg.configuration.get('depot_backend')
            local_backend = appcfg.configuration.get('depot_backend')

            if remote_backend == 'depot.io.local.LocalFileStorage':
                assert local_backend == remote_backend

                print("Fetching depot storage")

                local_depot = appcfg.configuration['depot_storage_path']
                remote_depot = remotecfg.configuration['depot_storage_path']

                remote_storage = os.path.join(remote_dir, remote_depot)
                local_storage = os.path.join('.', local_depot)

                if ':'.join((remote_storage, local_storage)) in fetched:
                    continue

                download_folder(remote_storage, local_storage)

                fetched.add(':'.join((remote_storage, local_storage)))

        if 'dsn' in remotecfg.configuration:
            assert 'dsn' in appcfg.configuration

            print("Fetching remote database")

            local_db = appcfg.configuration['dsn'].split('/')[-1]
            remote_db = remotecfg.configuration['dsn'].split('/')[-1]
            database_dump = subprocess.check_output([
                'ssh', server, '-C',
                (
                    "sudo -u postgres pg_dump "
                    "--no-owner --no-privileges --clean {}".format(
                        remote_db
                    )
                ),
            ])

            with open('dump.sql', 'w') as f:
                # add cascade to all drop schema lines
                # http://www.postgresql.org/message-id/50EC9574.9060500
                for line in database_dump.splitlines():
                    line = line.decode('utf-8')

                    if line.startswith('DROP SCHEMA'):
                        if 'CASCADE' not in line:
                            if 'DROP SCHEMA extensions' not in line:
                                line = line.replace(';', ' CASCADE;')

                    f.write(line + os.linesep)

            try:
                if platform.system() == 'Darwin':
                    subprocess.check_output([
                        "cat dump.sql | psql {}".format(local_db)
                    ], stderr=subprocess.STDOUT, shell=True)
                elif platform.system() == 'Linux':
                    subprocess.check_output([
                        (
                            "cat dump.sql "
                            "| sudo -u postgres psql -d {}".format(
                                local_db
                            )
                        )
                    ], stderr=subprocess.STDOUT, shell=True)
                else:
                    raise NotImplementedError
            except subprocess.CalledProcessError as e:
                lines = e.output.decode('utf-8').rstrip('\n').split('\n')

                if lines[-1].startswith('WARNING: errors ignored on restore:'):
                    pass
                else:
                    raise
            finally:
                subprocess.check_output(['rm', 'dump.sql'])


@cli.command(context_settings={'default_selector': '*'})
@click.option('--dry-run', default=False, is_flag=True,
              help="Do not write any changes into the database.")
@pass_group_context
def upgrade(group_context, dry_run):
    """ Upgrades all application instances of the given namespace(s). """

    modules = list(get_upgrade_modules())
    tasks = get_tasks()

    executed_raw_upgrades = set()

    basic_tasks = tuple((id, task) for id, task in tasks if not task.raw)
    raw_tasks = tuple((id, task) for id, task in tasks if task.raw)

    def on_success(task):
        print(click.style("* " + str(task.task_name), fg='green'))

    def on_fail(task):
        print(click.style("* " + str(task.task_name), fg='red'))

    def run_upgrade_runner(runner, *args):
        executed_tasks = runner.run_upgrade(*args)

        if executed_tasks:
            print("executed {} upgrade tasks".format(executed_tasks))
        else:
            print("no pending upgrade tasks found")

    def run_raw_upgrade(group_context, appcfg):
        if appcfg in executed_raw_upgrades:
            return

        executed_raw_upgrades.add(appcfg)

        title = "Running raw upgrade for {}".format(appcfg.path.lstrip('/'))
        print(click.style(title, underline=True))

        upgrade_runner = RawUpgradeRunner(
            tasks=raw_tasks,
            commit=not dry_run,
            on_task_success=on_success,
            on_task_fail=on_fail
        )

        run_upgrade_runner(
            upgrade_runner,
            appcfg.configuration['dsn'],
            group_context.available_schemas(appcfg),
        )

    def run_upgrade(request, app):
        title = "Running upgrade for {}".format(request.app.application_id)
        print(click.style(title, underline=True))

        upgrade_runner = UpgradeRunner(
            modules=modules,
            tasks=basic_tasks,
            commit=not dry_run,
            on_task_success=on_success,
            on_task_fail=on_fail
        )
        run_upgrade_runner(upgrade_runner, request)

    def upgrade_steps():
        if next((t for n, t in tasks if t.raw), False):
            yield run_raw_upgrade

        yield run_upgrade

    return tuple(upgrade_steps())



import html

from mistletoe import Document, HTMLRenderer
from onegov.core.html import sanitize_html


RENDERER_INSTANCES = {}


class HTMLRendererWithoutInlineHtml(HTMLRenderer):

    @staticmethod
    def render_html_block(token):
        return html.escape(token.content)

    @staticmethod
    def render_html_span(token):
        return html.escape(token.content)


def render_untrusted_markdown(markdown, cls=HTMLRendererWithoutInlineHtml):

    # use a global renderer instance, but only create it if used
    if cls not in RENDERER_INSTANCES:
        RENDERER_INSTANCES[cls] = cls()

    renderer = RENDERER_INSTANCES[cls]

    # mistletoe doesn't normalise line-endings, so we have to
    # see https://github.com/miyuchina/mistletoe/issues/64
    markdown = markdown.replace('\r\n', '\n')

    # render html
    html = renderer.render(Document(markdown))

    # clean it
    return sanitize_html(html)



import os
import psqlparse
import re

from onegov.core.cache import lru_cache
from onegov.core.orm import types as onegov_types
from sqlalchemy import text
from sqlalchemy import types as sqlalchemy_types
from uuid import uuid4


NESTED_TYPE = re.compile(r'(\w+)\((\w+)\)')


def as_selectable(query, alias=None):
    """ Takes a raw SQL query and turns it into a selectable SQLAlchemy
    expression using annotations in comments.

    Expects to find a SELECT statement like this::

        SELECT
            foo,  -- Text
            bar   -- Integer
        FROM
            foobar

    The so created selectable can be used by SQLAlchemy::

        from sqlalchemy import select

        query = as_selectable('''
            SELECT
                foo, -- Text
                bar  -- Integer
            FROM
                foobar
        ''')

        session.execute(select(query.c).where(query.c.foo == 'bar'))

    """

    # use the last statement if there are many
    statement = psqlparse.parse(query)[-1]

    # find the columns and use the comment to load types
    columns = {
        column: type_by_string(comment)
        for column, comment
        in column_names_with_comments(statement, query)
    }

    # an alias is required to reuse this query later
    alias = alias or ('q-' + uuid4().hex[:10])

    # turn the query into a selectable
    return text(query).columns(**columns).alias(alias)


def type_by_string(expression):
    nested = NESTED_TYPE.match(expression)

    if nested:
        name, nested = nested.groups()
        return type_by_string(name)(type_by_string(nested))

    return getattr(onegov_types, expression, None) \
        or getattr(sqlalchemy_types, expression)


@lru_cache(maxsize=64)
def as_selectable_from_path(path):
    alias = os.path.basename(path).split('.', 1)[0]

    with open(path, 'r') as f:
        return as_selectable(f.read(), alias=alias)


def column_names_with_comments(statement, query):
    for target in statement.target_list.targets:

        # expression
        if 'name' in target:
            column = target['name']

        # ordinary column
        elif 'val' in target and 'ColumnRef' in target['val']:
            column = target['val']['ColumnRef']['fields'][-1]['String']['str']

        else:
            raise NotImplementedError

        # find the next inline-comment
        location = target['location']

        at = query.find('--', location) + 2
        to = query.find('\n', at)

        assert at and to

        yield column, query[at:to].strip()



import pytest

from onegov.core.orm.abstract import (
    AdjacencyList,
    AdjacencyListCollection,
    MoveDirection,
    sort_siblings,
)


class FamilyMember(AdjacencyList):
    __tablename__ = 'familymembers'


class UnusedList(AdjacencyList):
    # this class leads to errors if we didn't abstract the adjacency list
    # properly (duplicate indexes and so on)
    __tablename__ = 'unusedlist'


class DeadFamilyMember(FamilyMember):
    __mapper_args__ = {'polymorphic_identity': 'dead'}


class FamilyMemberCollection(AdjacencyListCollection):
    __listclass__ = FamilyMember


def test_add(session):

    family = FamilyMemberCollection(session)

    adam = family.add_root('Adam')

    assert adam.name == 'adam'
    assert adam.title == 'Adam'
    assert adam.root is adam
    assert list(adam.ancestors) == []
    assert adam.path == 'adam'
    assert adam.absorb == 'adam'
    assert adam.type is None

    child = family.add(parent=adam, title='Cain')

    assert child.name == 'cain'
    assert child.title == 'Cain'
    assert child.root is adam
    assert list(child.ancestors) == [adam]
    assert child.path == 'adam/cain'
    assert child.absorb == 'adam/cain'

    with pytest.raises(AssertionError):
        # normalized names are enforced
        child.name = 'Cain'

    grandchild = family.add(parent=child, title='Enoch')
    assert list(grandchild.ancestors) == [adam, child]
    assert repr(grandchild) == "FamilyMember(name='enoch', id=3, parent_id=2)"


def test_add_unique_page(session):

    family = FamilyMemberCollection(session)
    r1 = family.add_root(title='Test')
    r2 = family.add_root(title='Test')
    r3 = family.add_root(title='Test')

    assert r1.name == 'test'
    assert r2.name == 'test-1'
    assert r3.name == 'test-2'

    c1 = family.add(parent=r1, title='Test')
    c2 = family.add(parent=r1, title='Test')
    c3 = family.add(parent=r1, title='Test')

    assert c1.name == 'test'
    assert c2.name == 'test-1'
    assert c3.name == 'test-2'


def test_add_or_get_page(session):

    family = FamilyMemberCollection(session)
    root = family.add_or_get_root(title='Wurzel', name='root')

    assert root.title == 'Wurzel'
    assert root.name == 'root'

    root = family.add_or_get_root(title='Wurzel', name='root')

    assert root.title == 'Wurzel'
    assert root.name == 'root'


def test_page_by_path(session):

    family = FamilyMemberCollection(session)

    news = family.add_root('News')
    assert family.by_path('/news') is news
    assert family.by_path('/news/') is news
    assert family.by_path('news/') is news
    assert family.by_path('news') is news
    assert family.by_path('/neues') is None

    toast = family.add(parent=news, title='Yeah Toast')
    assert family.by_path('/news/yeah-toast') is toast
    assert family.by_path('news/yeah-toast') is toast
    assert family.by_path('news/yeah-toast/') is toast
    assert family.by_path('/news/yeah-toast/') is toast
    assert family.by_path('/news/jah-toast') is None


def test_delete(session):
    family = FamilyMemberCollection(session)

    news = family.add_root('News')
    assert family.by_path('/news') is news

    family.delete(news)

    assert family.by_path('/news') is None


def test_polymorphic(session):
    family = FamilyMemberCollection(session)
    eve = family.add_root("Eve")

    member = family.add(parent=eve, title='Lara', type='dead')
    assert isinstance(member, DeadFamilyMember)

    member = family.add(parent=eve, title='Steve')
    assert not isinstance(member, DeadFamilyMember)

    with pytest.raises(AssertionError) as assertion_info:
        member = family.add(parent=eve, title='Don', type='other')

    assert "No such polymorphic_identity: other" in str(assertion_info.value)

    assert isinstance(family.by_path('/eve'), FamilyMember)
    assert isinstance(family.by_path('/eve/lara'), DeadFamilyMember)
    assert isinstance(family.by_path('/eve/steve'), FamilyMember)

    assert family.by_path('/eve/lara')
    assert family.by_path('/eve/lara', ensure_type='dead')
    assert not family.by_path('/eve/lara', ensure_type='missing')
    assert not family.by_path('/eve/inexistant', ensure_type='dead')


def test_move_root(session):

    family = FamilyMemberCollection(session)
    a = family.add_root("a")
    b = family.add_root("b")
    assert family.query(ordered=True).all() == [a, b]

    family.move_below(target=a, subject=b)
    assert family.query(ordered=True).all() == [a, b]

    family.move_above(target=a, subject=b)
    assert family.query(ordered=True).all() == [b, a]

    family.move_below(target=b, subject=a)
    assert family.query(ordered=True).all() == [b, a]

    family.move_above(target=b, subject=a)
    assert family.query(ordered=True).all() == [a, b]


def test_move(session):

    family = FamilyMemberCollection(session)

    root = family.add_root("root")
    a = family.add(parent=root, title='a')
    b = family.add(parent=root, title='b')

    assert a.siblings.all() == [a, b]

    family.move_below(target=a, subject=b)
    assert a.siblings.all() == [a, b]

    family.move_above(target=a, subject=b)
    assert a.siblings.all() == [b, a]

    family.move_below(target=b, subject=a)
    assert a.siblings.all() == [b, a]

    family.move_above(target=b, subject=a)
    assert a.siblings.all() == [a, b]


def test_move_keep_hierarchy(session):

    family = FamilyMemberCollection(session)

    a = family.add_root("a")
    b = family.add(parent=a, title='b')
    c = family.add(parent=b, title='c')

    for direction in MoveDirection:
        for target in (a, b, c):
            with pytest.raises(AssertionError):
                family.move(target=target, subject=a, direction=direction)

            with pytest.raises(AssertionError):
                family.move(target=target, subject=b, direction=direction)

            with pytest.raises(AssertionError):
                family.move(target=target, subject=c, direction=direction)


def test_add_sorted(session):
    family = FamilyMemberCollection(session)

    root = family.add_root("root")
    e = family.add(parent=root, title='e')
    c = family.add(parent=root, title='c')
    a = family.add(parent=root, title='a')

    query = family.query(ordered=True).filter(FamilyMember.parent_id != None)

    assert query.all() == [a, c, e]

    b = family.add(parent=root, title='b')

    assert query.all() == [a, b, c, e]

    # once the tree is unsorted, items are added at the bottom
    family.move_above(target=a, subject=e)
    assert query.all() == [e, a, b, c]

    d = family.add(parent=root, title='d')

    assert query.all() == [e, a, b, c, d]

    # until it is sorted again
    sort_siblings(a.siblings.all(), family.sort_key)
    assert query.all() == [a, b, c, d, e]

    aa = family.add(parent=root, title='aa')
    assert query.all() == [a, aa, b, c, d, e]


def test_change_title(session):
    family = FamilyMemberCollection(session)

    root = family.add_root("root")
    a = family.add(parent=root, title='a')
    b = family.add(parent=root, title='b')
    c = family.add(parent=root, title='c')

    assert a.siblings.all() == [a, b, c]

    b.title = 'd'

    assert a.siblings.all() == [a, c, b]


def test_change_title_unordered(session):
    family = FamilyMemberCollection(session)

    root = family.add_root("root")
    a = family.add(parent=root, title='a', order=3)
    b = family.add(parent=root, title='b', order=1)
    c = family.add(parent=root, title='c', order=2)

    assert a.siblings.all() == [b, c, a]

    b.title = 'z'

    assert a.siblings.all() == [b, c, a]



from sqlalchemy_utils import QueryChain as QueryChainBase


class QueryChain(QueryChainBase):
    """ Extends SQLAlchemy Utils' QueryChain with some extra methods. """

    def slice(self, start, end):
        return self[start:end]

    def first(self):
        return next((o for o in self), None)

    def all(self):
        return tuple(self)



""" Provides caching methods for onegov.core.

Onegov.core uses dogpile for caching:
`<https://dogpilecache.readthedocs.org/>`_

Unlike dogpile onegov.core does not provide a global region however.
The cache is available through the app::

    request.app.cache.set('key', 'value')

Global caches in a multi-tennant application are a security vulnerability
waiting to be discovered. Therefore we do not do that!

This means that this won't be available::

    from x import cache
    @cache.cache_on_arguments()
    def my_function():
        return 'foobar'

Syntactic sugar like this will be provided through decorators inside this
module in the future. For example, we could write one that is usable
on all morepath views::

    @App.view(...)
    @cache.view()
    def my_view():
        return '<html>...'

But no such method exists yet.

Currently there is one cache per app that never expires (though values will
eventually be discarded by redis if the cache is full).

"""

import dill

from fastcache import clru_cache as lru_cache  # noqa
from dogpile.cache import make_region
from dogpile.cache.api import NO_VALUE
from dogpile.cache.proxy import ProxyBackend
from redis import ConnectionPool


@lru_cache(maxsize=1024)
def get(namespace, expiration_time, redis_url):

    def key_mangler(key):
        return f'{namespace}:{key}'.encode('utf-8')

    return make_region(key_mangler=key_mangler).configure(
        'dogpile.cache.redis',
        arguments={
            'url': redis_url,
            'redis_expiration_time': expiration_time + 1,
            'connection_pool': get_pool(redis_url)
        },
        wrap=(DillSerialized, )
    )


@lru_cache(maxsize=16)
def get_pool(redis_url):
    return ConnectionPool.from_url(redis_url)


class DillSerialized(ProxyBackend):
    """ A proxy backend that pickles all non-byte values using dill. """

    def serialize(self, value):
        if isinstance(value, bytes):
            return value

        return dill.dumps(value, recurse=True)

    def deserialize(self, value):
        if value is NO_VALUE:
            return value

        return dill.loads(value)

    def get(self, key):
        return self.deserialize(self.proxied.get(key))

    def set(self, key, value):
        return self.proxied.set(key, self.serialize(value))

    def get_multi(self, keys):
        return [self.deserialize(v) for v in self.proxied.get_multi(keys)]

    def set_multi(self, mapping):
        mapping = {k: self.serialize(v) for k, v in mapping.items()}
        self.proxied.set_multi(mapping)



import os
import os.path
import polib
import pytest

from gettext import GNUTranslations
from onegov.core import i18n, utils
from webob import Request
from wtforms import Label

# compatibility shim for webob 1.8 before its release
try:
    from webob.acceptparse import create_accept_language_header
except ImportError:
    from webob.acceptparse import Accept as create_accept_language_header


def test_pofiles(temporary_directory):
    os.makedirs(os.path.join(temporary_directory, 'de/LC_MESSAGES'))
    utils.touch(
        os.path.join(temporary_directory, 'de/LC_MESSAGES/onegov.test.po'))

    result = list(i18n.pofiles(temporary_directory))
    assert result[0][0] == 'de'
    assert result[0][1].endswith('onegov.test.po')


def test_merge_translations():
    t1 = GNUTranslations()
    t1._catalog = {
        'We are here': 'Hier sind wir',
        'Welcome': 'Willkommen'
    }

    t2 = GNUTranslations()
    t2._catalog = {
        'We are here': 'Wir sind hier',
        'Hi': 'Hallo'
    }

    t3 = GNUTranslations()
    t3._catalog = {
        'We are here': 'Sind wir hier',
        'Morning': 'Moin'
    }

    with pytest.raises(AssertionError):
        i18n.merge([])

    with pytest.raises(AssertionError):
        i18n.merge([i18n.clone(t1)])

    assert i18n.merge([i18n.clone(t1), i18n.clone(t2)])._catalog == {
        'We are here': 'Hier sind wir',
        'Welcome': 'Willkommen',
        'Hi': 'Hallo'
    }

    assert i18n.merge([i18n.clone(t2), i18n.clone(t3)])._catalog == {
        'We are here': 'Wir sind hier',
        'Morning': 'Moin',
        'Hi': 'Hallo'
    }

    assert i18n.merge(
        [i18n.clone(t1), i18n.clone(t2), i18n.clone(t3)])._catalog == {
        'We are here': 'Hier sind wir',
        'Welcome': 'Willkommen',
        'Morning': 'Moin',
        'Hi': 'Hallo'
    }


def test_get_translations(temporary_directory):
    os.makedirs(os.path.join(temporary_directory, '1/de/LC_MESSAGES'))
    po = polib.POFile()
    po.append(polib.POEntry(
        msgid='Welcome',
        msgstr='Willkommen'
    ))
    po.save(os.path.join(
        temporary_directory, '1/de/LC_MESSAGES/onegov.test.po'))

    os.makedirs(os.path.join(temporary_directory, '2/de/LC_MESSAGES'))
    po = polib.POFile()
    po.append(polib.POEntry(
        msgid='Morning',
        msgstr='Moin'
    ))
    po.save(os.path.join(
        temporary_directory, '2/de/LC_MESSAGES/onegov.test.po'))

    os.makedirs(os.path.join(temporary_directory, '2/fr/LC_MESSAGES'))
    po = polib.POFile()
    po.append(polib.POEntry(
        msgid='Welcome',
        msgstr='Bienvenue'
    ))
    po.save(os.path.join(
        temporary_directory, '2/fr/LC_MESSAGES/onegov.test.po'))

    translations = i18n.get_translations([
        temporary_directory + '/1',
        temporary_directory + '/2'
    ])
    assert list(sorted(translations.keys())) == list(sorted(['de', 'fr']))

    translations['de'].gettext('Welcome') == 'Willkommen'
    translations['de'].gettext('Morning') == 'Moin'
    translations['fr'].gettext('Welcome') == 'Bienvenue'

    translations = i18n.get_translations([
        temporary_directory + '/1',
    ])
    assert list(sorted(translations.keys())) == list(sorted(['de']))
    translations['de'].gettext('Welcome') == 'Willkommen'
    translations['de'].gettext('Morning') == 'Morning'

    translations = i18n.get_translations([
        temporary_directory + '/2',
    ])
    assert list(sorted(translations.keys())) == list(sorted(['de', 'fr']))
    translations['de'].gettext('Welcome') == 'Welcome'
    translations['de'].gettext('Morning') == 'Moin'
    translations['fr'].gettext('Welcome') == 'Bienvenue'


def test_default_locale_negotiator():

    negotiate = i18n.default_locale_negotiator
    request = Request.blank('/')

    request.accept_language = create_accept_language_header('fr')
    assert negotiate(['en', 'de'], request) is None

    request.accept_language = create_accept_language_header('en')
    assert negotiate(['en', 'de'], request) == 'en'

    request.accept_language = create_accept_language_header('de')
    assert negotiate(['en', 'de'], request) == 'de'

    request.accept_language = create_accept_language_header('de')
    request.cookies['locale'] = 'en'
    assert negotiate(['en', 'de'], request) == 'en'


def test_get_translation_bound_form():

    class MockTranslation(object):

        _fallback = None

        def add_fallback(self, translation):
            self._fallback = translation

        def gettext(self, text):
            return text[::-1]  # inverse the string

    default = MockTranslation()
    translate = MockTranslation()

    class MetaMeta(object):

        def render_field(self, field, render_kw):
            return field.label.text

    class MockMeta(MetaMeta):

        def get_translations(self, form):
            return default

    class MockForm(object):

        Meta = MockMeta

    class MockField(object):

        def __init__(self, label):
            self.label = Label('x', label)

    form_class = i18n.get_translation_bound_form(MockForm, translate)

    assert form_class.Meta().get_translations(None) is translate
    assert form_class.Meta().get_translations(None)._fallback is default

    meta = form_class.Meta()
    meta.get_translations(None)
    assert meta.render_field(MockField('Yes'), {}) == 'seY'



import logging
import warnings

log = logging.getLogger('onegov.core')  # noqa
log.addHandler(logging.NullHandler())  # noqa

ignored_warnings = (
    # we will keep using psycopg2 instead of psycogp2-binary
    "The psycopg2 wheel package will be renamed from release 2.8",
)

for message in ignored_warnings:
    warnings.filterwarnings("ignore", message=message)

from onegov.core.framework import Framework # noqa
from onegov.core.filestorage import get_filestorage_file # noqa

# include the filters module so they get picked up by webassets
from onegov.core import filters  # noqa

__all__ = ['Framework', 'log']



import pytest

from dogpile.cache import make_region
from dogpile.cache.api import NO_VALUE
from onegov.core.browser_session import BrowserSession
from onegov.core import cache


def test_browser_session_mangle():
    session = BrowserSession({}, 'token')
    assert session._cache.mangle('test')\
        == 'e352b6fc45b2bc144082d507d6a51faec0bbeab5313974f7:test'

    with pytest.raises(AssertionError):
        session._cache.mangle('')

    with pytest.raises(AssertionError):
        session._cache.mangle(None)


def test_browser_session_cache():
    cache = make_region().configure('dogpile.cache.memory')
    session = BrowserSession(cache, 'token')
    key = 'e352b6fc45b2bc144082d507d6a51faec0bbeab5313974f7:name'

    assert not session.has('name')
    assert 'name' not in session

    with pytest.raises(AttributeError):
        session.name

    with pytest.raises(KeyError):
        session['name']

    assert cache.get(key) is NO_VALUE

    session.name = 'test'
    assert session.has('name')
    assert 'name' in session
    assert session.name == 'test'
    assert session['name'] == 'test'
    assert cache.get(key) == 'test'

    del session.name
    assert not session.has('name')
    assert 'name' not in session

    with pytest.raises(AttributeError):
        session.name

    with pytest.raises(KeyError):
        session['name']

    assert cache.get(key) is NO_VALUE


def test_browser_session_cache_prefix():
    cache = make_region().configure('dogpile.cache.memory')

    session = BrowserSession(cache, 'foo')
    session.name = 'test'
    assert session.name == 'test'

    session = BrowserSession(cache, 'foo')
    assert session.name == 'test'

    session = BrowserSession(cache, 'bar')
    with pytest.raises(AttributeError):
        session.name

    session = BrowserSession(cache, 'bar')
    with pytest.raises(AttributeError):
        session.name


def test_browser_session_count(redis_url):
    session = BrowserSession(cache.get('sessions', 60, redis_url), 'token')

    assert session.count() == 0
    session['foo'] = 'bar'

    assert session.count() == 1
    session['foo'] = 'baz'

    assert session.count() == 1
    del session['foo']

    assert session.count() == 0

    session['asdf'] = 'qwerty'
    assert session.count() == 1

    session.flush()
    assert session.count() == 0

    session.flush()
    assert session.count() == 0



import os
import tempfile
import transaction

from onegov.core import log


class MailDataManager(object):
    """ Takes a postman and an envelope and sends it when the transaction
    is commited.

    Since we can't really know if a mail can be sent until it happens, we
    simply log an exception if the sending failed.

    """

    transaction_manager = transaction.manager

    def __init__(self, postman, envelope):
        self.postman = postman
        self.envelope = envelope

    @classmethod
    def send_email(cls, postman, envelope):
        transaction.get().join(cls(postman, envelope))

    def sortKey(self):
        return 'mails'

    def bind_connection(self, transaction, connection):
        assert 'mail_connection' not in transaction.extension

        def after_commit_hook(*args):
            connection.quit()

        transaction.addAfterCommitHook(after_commit_hook)
        transaction.extension['mail_connection'] = connection

    def open_connection(self):
        connection = self.postman.transport(
            self.postman.host,
            self.postman.port,
            **self.postman.options
        )

        connection.ehlo()

        for item in self.postman.middlewares:
            item(connection)

        return connection

    def commit(self, transaction):
        if 'mail_connection' not in transaction.extension:
            self.bind_connection(transaction, self.open_connection())

        try:
            self.postman.deliver(
                transaction.extension['mail_connection'],
                self.envelope
            )
        except Exception:
            log.exception("Failed to send e-mail")

    def abort(self, transaction):
        pass

    def tpc_vote(self, transaction):
        pass

    def tpc_abort(self, transaction):
        pass

    def tpc_begin(self, transaction):
        pass

    def tpc_finish(self, transaction):
        pass


class FileDataManager(object):

    """ Writes a file when the transaction is commited. """

    transaction_manager = transaction.manager

    def __init__(self, data, path):
        self.data = data
        self.path = path

    @classmethod
    def write_file(cls, data, path):
        transaction.get().join(cls(data, path))

    def sortKey(self):
        return 'files'

    def commit(self, transaction):
        with tempfile.NamedTemporaryFile(delete=False) as temp:
            self.tempfn = temp.name
            temp.write(self.data)

    def abort(self, transaction):
        pass

    def tpc_vote(self, transaction):
        if not os.path.exists(self.tempfn):
            raise ValueError('%s doesnt exist' % self.tempfn)
        if os.path.exists(self.path):
            raise ValueError('file already exists')

    def tpc_abort(self, transaction):
        try:
            os.remove(self.tempfn)
        except OSError:
            pass

    def tpc_begin(self, transaction):
        pass

    def tpc_finish(self, transaction):
        os.rename(self.tempfn, self.path)



""" Provides a simple and mostly transparent way of defining orm-cached
properties on the application class.

For example::

    from onegov.core import Framework
    from onegov.core.orm import orm_cached

    class App(Framework):

        @orm_cached(policy='on-table-change:users')
        def users(self):
            # ... fetch users from database

Properties defined in this way are accessible through the instance::

    app.users

If there are any changes to the users table, the cache is removed. Since the
cache is usually a shared redis instance, this works for multiple processes.

"""

import inspect
import sqlalchemy

from sqlalchemy.orm.query import Query


class OrmCacheApp(object):
    """ Integrates the orm cache handling into the application
    (i.e. :class:`onegov.core.framework.Framework').

    In addition, the application needs to call :meth:`setup_orm_cache` inside
    of `:meth:onegov.server.application.Application.set_application_id` to
    enable the cache evicition mechanism.

    """

    def configure_orm_cache(self, **cfg):
        self.is_orm_cache_setup = getattr(self, 'is_orm_cache_setup', False)

    def setup_orm_cache(self):
        """ Sets up the event handlers for the change-detection. """

        assert not self.is_orm_cache_setup

        for descriptor in self.orm_cache_descriptors:
            self.session_manager.on_insert.connect(
                self.descriptor_bound_orm_change_handler(descriptor),
                weak=False)
            self.session_manager.on_update.connect(
                self.descriptor_bound_orm_change_handler(descriptor),
                weak=False)
            self.session_manager.on_delete.connect(
                self.descriptor_bound_orm_change_handler(descriptor),
                weak=False)

        self.is_orm_cache_setup = True

    def descriptor_bound_orm_change_handler(self, descriptor):
        """ Listens to changes to the database and evicts the cache if the
        policy demands it. Available policies:

        * policy='on-table-change:table': clears the cache if there's a change
          on the given table

        * policy=lambda obj: ...: clears the cache if the given policy function
          returns true (it receives the object which has changed)

        """

        def handle_orm_change(schema, obj):

            if callable(descriptor.cache_policy):
                dirty = descriptor.cache_policy(obj)

            elif descriptor.cache_policy.startswith('on-table-change'):
                tablename = descriptor.cache_policy.split(':')[-1]
                dirty = obj.__class__.__tablename__ == tablename

            else:
                raise NotImplementedError()

            if dirty:
                # Two circumstances ensure that only the cache of the current
                # schema is evicted:
                #
                # 1. There's a current schema set when the event is fired.
                # 2. The current cache is bound to the current schema.
                #
                # Still, trust but verify:
                assert self.schema == schema
                self.cache.delete(descriptor.cache_key)

                if descriptor.cache_key in self.request_cache:
                    del self.request_cache[descriptor.cache_key]

        return handle_orm_change

    @property
    def orm_cache_descriptors(self):
        """ Yields all orm cache descriptors installed on the class. """

        for member_name, member in inspect.getmembers(self.__class__):
            if isinstance(member, OrmCacheDescriptor):
                yield member


class OrmCacheDescriptor(object):
    """ The descriptor implements the protocol for fetching the objects
    either from cache or from the database (through the handler).

    """

    def __init__(self, cache_policy, creator):
        self.cache_policy = cache_policy
        self.cache_key = creator.__qualname__
        self.creator = creator

    def create(self, instance):
        """ Uses the creator to load the object to be cached.

        Since the return value of the creator might not be something we want
        to cache, this function will turn some return values into something
        more useful (e.g. queries are completely fetched).

        """

        result = self.creator(instance)

        if isinstance(result, Query):
            result = tuple(result)

        return result

    def merge(self, session, obj):
        """ Merges the given obj into the given session, *if* this is possible.

        That is it acts like more forgiving session.merge().

        """
        if self.requires_merge(obj):
            obj = session.merge(obj, load=False)
            obj.is_cached = True

        return obj

    def requires_merge(self, obj):
        """ Returns true if the given object requires a merge, which is the
        case if the object is detached.

        """

        # no need for an expensive sqlalchemy.inspect call for these
        if isinstance(obj, (int, str, bool, float, tuple, list, dict, set)):
            return False

        info = sqlalchemy.inspect(obj, raiseerr=False)

        if not info:
            return False

        return info.detached

    def load(self, app):
        """ Loads the object from the database or cache. """

        session = app.session()

        # before accessing any cached values we need to make sure that all
        # pending changes are properly flushed -> this leads to some extra cpu
        # cycles spent but eliminates the chance of accessing a stale entry
        # after a change
        if session.dirty:
            session.flush()

        # we use a secondary request cache for even more lookup speed and to
        # make sure that inside a request we always get the exact same instance
        # (otherwise we don't see changes reflected)
        if self.cache_key in app.request_cache:

            # it is possible for objects in the request cache to become
            # detached - in this case we need to merge them again
            # (the merge function only does this if necessary)
            return self.merge(session, app.request_cache[self.cache_key])

        else:
            obj = app.cache.get_or_create(
                key=self.cache_key,
                creator=lambda: self.create(app)
            )

        # named tuples
        if isinstance(obj, tuple) and hasattr(obj.__class__, '_make'):
            obj = obj._make(self.merge(session, o) for o in obj)

        # lists (we can save some memory here)
        elif isinstance(obj, list):
            for ix, o in enumerate(obj):
                obj[ix] = self.merge(session, o)

        # generic iterables
        elif isinstance(obj, (tuple, set)):
            obj = obj.__class__(self.merge(session, o) for o in obj)

        # generic objects
        else:
            obj = self.merge(session, obj)

        app.request_cache[self.cache_key] = obj

        return obj

    def __get__(self, instance, owner):
        """ Handles the object/cache access. """

        if instance is None:
            return self

        return self.load(instance)


def orm_cached(policy):
    """ The decorator use to setup the cache descriptor.

    See the :mod:`onegov.core.orm.cache` docs for usage.

    """

    def orm_cache_decorator(fn):
        return OrmCacheDescriptor(policy, fn)
    return orm_cache_decorator



from contextlib import suppress
from dogpile.cache.api import NO_VALUE
from hashlib import blake2b


class Prefixed(object):

    def __init__(self, prefix, cache):
        assert len(prefix) >= 24

        self.prefix = prefix
        self.cache = cache

    def mangle(self, key):
        assert key
        return f'{self.prefix}:{key}'

    def get(self, key):
        return self.cache.get(self.mangle(key))

    def set(self, key, value):
        return self.cache.set(self.mangle(key), value)

    def delete(self, key):
        return self.cache.delete(self.mangle(key))

    def count(self):
        # note, this cannot be used in a Redis cluster - if we use that
        # we have to keep track of all keys separately
        return self.cache.backend.proxied.client.eval("""
            return #redis.pcall('keys', ARGV[1])
        """, 0, f'*:{self.prefix}:*')

    def flush(self):
        # note, this cannot be used in a Redis cluster - if we use that
        # we have to keep track of all keys separately
        return self.cache.backend.proxied.client.eval("""
            if #redis.pcall('keys', ARGV[1]) > 0 then
                return redis.pcall('del', unpack(redis.call('keys', ARGV[1])))
            end
        """, 0, f'*:{self.prefix}:*')


class BrowserSession(object):
    """ A session bound to a token (session_id cookie). Used to store data
    about a client on the server.

    Instances should be called ``browser_session`` to make sure we don't
    confuse this with the orm sessions.

    Used by :class:`onegov.core.request.CoreRequest`.

    Example:

        browser_session = request.browser_session
        browser_session.name = 'Foo'
        assert client.session.name == 'Foo'
        del client.session.name

    This class can act like an object, through attribute access, or like a
    dict, through square brackets. Whatever you prefer.

    """

    def __init__(self, cache, token, on_dirty=lambda session, token: None):

        # make it impossible to get the session token through key listing
        prefix = blake2b(token.encode('utf-8'), digest_size=24).hexdigest()
        self._cache = Prefixed(prefix=prefix, cache=cache)
        self._token = token

        self._is_dirty = False
        self._on_dirty = on_dirty

    def has(self, name):
        return self._cache.get(name) is not NO_VALUE

    def flush(self):
        self._cache.flush()
        self.mark_as_dirty()

    def count(self):
        return self._cache.count()

    def pop(self, name, default=None):
        """ Returns the value for the given name, removing the value in
        the process.

        """
        value = self.get(name, default=default)

        # we can run into a race condition here when two requests come in
        # simultaneously - one request will get the value and delete it, the
        # other will get the value and fail with an error when trying to
        # delete it
        #
        # we can be pragmatic - if it's gone, it doesn't need to be deleted
        with suppress(AttributeError):
            delattr(self, name)

        return value

    def mark_as_dirty(self):
        if not self._is_dirty:
            self._on_dirty(self, self._token)
            self._is_dirty = True

    def get(self, name, default=None):
        result = self._cache.get(name)

        if result is NO_VALUE:
            return default

        return result

    def __getitem__(self, name, *args):
        result = self._cache.get(name)

        if result is NO_VALUE:
            raise KeyError

        return result

    def __getattr__(self, name):
        result = self._cache.get(name)

        if result is NO_VALUE:
            raise AttributeError

        return result

    def __setattr__(self, name, value):
        if name.startswith('_'):
            return super().__setattr__(name, value)

        self._cache.set(name, value)
        self.mark_as_dirty()

    def __delattr__(self, name):
        if name.startswith('_'):
            return super().__delattr__(name)

        self._cache.delete(name)
        self.mark_as_dirty()

    __setitem__ = __setattr__
    __delitem__ = __delattr__
    __delattr__ = __delattr__
    __contains__ = has



import mailthon.headers as headers
import os.path

from email.mime.multipart import MIMEMultipart
from mailbox import Maildir, MaildirMessage
from mailthon.enclosure import HTML, PlainText, Attachment, Enclosure
from mailthon.envelope import Envelope as BaseEnvelope
from mailthon.postman import Postman
from onegov.core.html import html_to_text


def email(sender=None, receivers=(), cc=(), bcc=(),
          subject=None, content=None, encoding='utf8',
          attachments=(), category='marketing', plaintext=None):
    """
    Creates an Envelope object with a HTML *content*, as well as a *plaintext*
    alternative generated from the HTML content.

    :param content: HTML content.
    :param encoding: Encoding of the email.
    :param attachments: Either a list of :class:`mailthon.enclosure.Enclosure`
        or a list of filenames to attach to the email.

    Note: this is basically a copy of :func:`mailthon.api.email`, though it
    adds the plaintext alternative.
    """

    if content and not plaintext:
        # turn the html email into a plaintext representation
        # this leads to a lower spam rating
        plaintext = html_to_text(content)

    headers_ = [
        headers.subject(subject),
        headers.sender(sender),
        headers.to(*receivers),
        headers.cc(*cc),
        headers.bcc(*bcc),
        headers.date(),
        headers.message_id(),
        ('X-Category', category),
    ]

    # According to RFC 2046, the last part of a multipart message, in this
    # case the HTML message, is best and preferred
    enclosure = [PlainText(plaintext, encoding), HTML(content, encoding)]

    if not attachments:
        # simple 'mime-multipart/alternative'
        envelope = Envelope(headers=headers_, enclosure=enclosure)
    else:
        # 'multipart/mixed' with 'mime-multipart/alternative' and attachments
        enclosure = [Envelope(headers=[], enclosure=enclosure)]
        enclosure.extend(
            k if isinstance(k, Enclosure) else Attachment(k)
            for k in attachments
        )
        envelope = BaseEnvelope(headers=headers_, enclosure=enclosure)

    return envelope


class Envelope(BaseEnvelope):
    """ Changes the mailthon envelope to use mime-multipart/alternative.

    """

    def mime(self):
        mime = MIMEMultipart('alternative')
        for item in self.enclosure:
            mime.attach(item.mime())
        self.headers.prepare(mime)
        return mime


class MaildirTransport(object):
    """ A transport that pretends to be like python's smtplib.SMTP transport,
    but actually just stores files into a maildir.

    """

    def __init__(self, host, port, maildir):
        # host and port exist for mailthon compatibility, they are ignored
        self.maildir = Maildir(maildir, create=True)

        for d in (os.path.join(maildir, d) for d in ('new', 'cur', 'tmp')):
            if not os.path.exists(d):
                os.makedirs(d)

    def sendmail(self, from_addr, to_addrs, message):
        self.maildir.add(MaildirMessage(message))

        return {}

    def noop(self):
        return 250, ""

    def ehlo(self):
        pass

    def quit(self):
        pass


class MaildirPostman(Postman):
    """ A mailthon postman that stores mails to a maildir. """
    transport = MaildirTransport
